/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file
 * overall include file for OBS.
 */

#ifndef OBS_H_
#define OBS_H_

/**@mainpage OBS - object stack/slab allocator
 *
 * @section intro introduction
 * @em OBS is implementation of two kinds of pool allocator,
 * object stack and slab.
 *
 * Object stack allocator is similar to GNU obstack.
 * Compared with GNU obstack, OBS stack allocator has the following traits:
 *
 *  - (L)GPL-free (BSD license) (*)
 *  - using mmap-based page allocator
 *  - rather better interface
 *  - can truncate not only stack top but also stack bottom
 *
 * @note (*) Unfortunately, except for the build system - thankfully using
 * GNU autoconf/automake/libtool, although these do not affect the
 * binary form of OBS library with (L)GPL.
 *
 * Slab allocator, which is normally used for kernel allocator,
 * is another allocator OBS including.  Each slab allocator instance keeps
 * one or more objects which are the same type.  Once an object is
 * allocated, the allocator caches the object for a while even after
 * the object is returned to the allocator.
 *
 *
 * @section download how to get OBS
 * @em OBS can be downloaded from
 * <A HREF="http://www.bsdclub.org/~tshiozak/obs/dist/">
 *   http://www.bsdclub.org/~tshiozak/obs/dist/
 * </A>
 *
 *
 * @section stack_concept concepts of stack allocator
 * Although there are all sorts of pool allocator, they basically have
 * characteristics as follows:
 *  - supporting fast deallocation all objects in each pool
 *  - supporting successively allocation of objects
 *  - per-object deallocation is impossible or strongly restricted
 *
 * Since the heaviest process on allocator is sometimes deallocation and
 * its accompanying processes, overall processes can be light by simplifying
 * deallocation.  Additionally, it slightly prevents memory leaks because of
 * collective deallocation.  It however apts to waste memory, because
 * per-object deallocation is rectricted.
 *
 * OBS stack, and also its ancestor GNU obstack, can be used like stack, so
 *  - can resize only the object on the stack top (growth of stack), and
 *  - can collectively deallocate objects which were allocated after
 *    a specific object.
 *
 * Unlike general stack, OBS stack and GNU obstack manage memory dividing into
 * multiple blocks.
 *
 * @subsection stack_stack stack
 * Multiple stack instances can be used at the same time.
 * Each stack is represented by ::OBS_stack
 * structure.  ::OBS_stack structure is a opaque (meaning users cannot access
 * its members - gentlemen's agreement) but completely defined
 * (meaning users can define its variable).
 * Typical usage of ::OBS_stack is as follows:
@verbatim
void
foo()
{
	OBS_status ret;
	OBS_stack stack;

	// initialize stack
	ret = obs_stack_init(&stack, 1, 0, OBS_ALIGN_1);
	if (ret)
		error process...;
	normal processes...
	// dispose stack
	obs_stack_uninit(&stack);
}
@endverbatim
 *
 *
 * @subsection stack_top stack top object
 * @em stack @em top @em object is the memory area which is the most recently
 * allocated one.  This is resizable until it is @em fixated explicitly,
 * and may be relocated on resizing.
 * In this document, stack top object is also called
 * "non-fixated stack top object" to clarify such attribute, although
 * it is redundant because stack top object is always non-fixated.
 *
 * Stack top object can takes one of the following three states:
 * invalid, empty and normal.
 * Just after initialization, the stack keeps an invalid stack top object.
 * The head pointer of the invalid stack top object takes NULL, and
 * size is undefined.
 * Just after calling obs_stack_fixate_top() function,
 * the stack keeps an empty stack top object.
 * The head pointer of the empty stack top object takes some valid pointer,
 * and size is zero.
 * To empty the stack, obs_stack_revoke_top() function can be used.
 * Invalid and zero stack top object are neither valid object.
 * To make an valid stack top object, obs_stack_alloc() or the like can
 * be used.
 *
 * In order to fixate a valid stack top object, obs_stack_fixate_top()
 * function can be used. obs_stack_alloc() function can also be used,
 * it fixates the current stack top,
 * and then creates a new non-fixed stack top object.
 * Once an object is fixated, it is not relocated until its deallocation.
 *
 *
 * @subsection stack_alloc_resize allocation, resize and expanse stack top.
 * To allocate, resize or expand stack, obs_stack_alloc() function,
 * obs_stack_resize() function or obs_stack_expand() function can be used
 * respectively.
 *
 * obs_stack_alloc() function is used to allocate a new stack top object.
 * To begin with, it fixates the current stack top object if it is
 * valid.  Next, a new stack top object whose size is specified by
 * the argument is created.
 *
 * obs_stack_resize() function is used to resize the current stack top object.
 * it is equivalent to obs_stack_alloc() if the current stack top object
 * is invalid or zero.  Otherwise, it resizes the stack top to the size
 * which is specified by the argument.
 *
 * obs_stack_expand() function is the same as obs_stack_resize() function,
 * except the size is specified as increment.
 *
 * obs_stack_resize() and obs_stack_expand() may relocate the
 * stack top object.  Shrinking the stack top, however, does not cause
 * relocation.
 *
 *
 * @subsection stack_append append bytes
 * To append bytes, obs_stack_append() function, obs_stack_append_region()
 * function and obs_stack_append_1() function can be used.
 *
 * obs_stack_append() function is like obs_stack_expand().
 * After expanse, the function copies the bytes specified by the argument.
 *
 * obs_stack_append() function is the same as obs_stack_append() function,
 * except for the form of the argument;  ::OBS_region is used instead of
 * raw void pointer and size.
 *
 * obs_stack_append_1() function is used to append a byte.
 *
 * These functions may cause relocation.
 *
 *
 * @subsection stack_reserve reserve
 * obs_stack_reserve() function is used to reserve area for the current
 * stack top object.  This is like obs_stack_resize(), but does not
 * change the real size of the stack top and just allocate memory block.
 * However, this may causes relocation for the stack top object.
 * This function can be used to improve performance on gradual allocation.
 *
 * When the enough memory is reserved, obs_stack_resize_fast() function,
 * obs_stack_expand_fast() function, obs_stack_append_fast() function,
 * obs_stack_append_region_fast() function and obs_stack_append_1_fast()
 * function can be used instead of the respective non-fast functions.
 * These functions never check the size, thus there need to be used carefully.
 *
 *
 * @subsection stack_truncation stack truncation
 * OBS supports truncation like stack.  On general stacks, the program
 * can save the stack pointer at some point, can push some objects on the
 * stack, and then can free all objects allocated after the stack pointer
 * saving by restoring the stack pointer.  On OBS, similar deallocation
 * can be done by specifying a pointer pointing object which is allocated
 * in the past, then all objects which are allocated after the specified
 * object, including specified one, are deallocated.
 * A pointer which points non-fixated stack top object or which points
 * the middle of a object can be also specified.  In the latter case,
 * the object of which middle is specified is divided at the pointer,
 * the latter half is deallocated, and the former half is keeped.
 * On OBS, obs_stack_truncate_after() function can be used to such truncation.
 *
 * GNU obstack and general stacks support only such truncation,
 * which deallocates towards stack top.   Unlike them, OBS also supports
 * deallocation towards stack bottom, like queue.
 * obs_stack_truncate_before() function can be used to such sort of truncation.
 *
 *
 * @subsection stack_alignment alignment
 * An alignment value to be used for allocating stack top object can be
 * specified by using obs_stack_set_alignment() function.
 * The alignment argument of this function is the form of an amount of
 * bit shift.  Namely, 0 means 1 byte alignment (1<<0), 1 means 2 bytes
 * alignment (1<<1), 2 means 4 bytes alignment(1<<2) and so on.
 * To get the max alignment value, obs_stack_get_max_alignment() function
 * can be used.
 *
 *
 * @subsection stack_thread thread safety
 * ::OBS_stack structure cannot be accessed at the same time.
 * It is thread safe only if an ::OBS_stack is keeped exclusive among all
 * threads, or per-thread ::OBS_stack structures are used.
 *
 *
 * @section slab_concepts concepts of slab allocator
 * Slab allocator is another allocator OBS is providing.
 * Slab is fit if a number of objects which are the same type are
 * allocated and deallocated repetitiously.  However, an allocator
 * instance needs to be prepared every type of objects.
 *
 *
 * @subsection slab_usage basic usage of slab allocator
 * Each slab allocator is represented by ::OBS_slab structure.
 * ::OBS_slab structure is a opaque but completely defined.
 * Typical usage of ::OBS_slab is as follows:
@verbatim
void
foo()
 {
	OBS_status ret;
	OBS_slab slab;
	struct object *obj1, *obj2;

	// initialize slab
	ret = obs_slab_init(&slab, sizeof (struct object), 0,
			    OBS_ALIGN_POINTER, NULL, NULL);
	if (ret)
		error process...;

	// get two objects
	ret = obs_slab_get(&slab, (void **)&obj1);
	if (ret)
		error process...;
	ret = obs_slab_get(&slab, (void **)&obj2);
	if (ret)
		error process...;

	// put one of them back
	ret = obs_slab_put(&slab, (void **)&obj1);
	if (ret)
		error process...;

	// dispose slab. all objects are freed.
	obs_slab_uninit(&slab);
}
@endverbatim
 *
 *
 *
 * @subsection slab_object_cache slab as object cache
 * Slab allocator has a side as an object cache. Objects which have become
 * unnecessary are cached by the allocator a while, and reused for the
 * next allocation.  Constructor and destructor of objects is called
 * by allocator as the need arises.
 *
 * All objects belonging to the slab allocator has one of the three states:
 * "uninited", "unused" and "inuse".
 * The first and second state is also called "free" state.
 *
 *
 * @subsection slab_gc garbage collection
 * To manage objects, the allocator puts several objects together into
 * a memory chunk called "slab".
 * Thus, the slab allocator can return the slabs to the backend page
 * allocator iff there are slabs which keeps no objects in use.
 * Such return may be invoked automatically or manually.
 * By default, automatic garbage collection is disabled.
 *
 * Manual garbage collection is invoked by calling obs_slab_run_gc() function.
 * Automatic garbage collection is set up by calling obs_slab_set_water_mark()
 * function.  See the description about respective functions for the details.
 *
 *
 * @subsection stack_alignment alignment
 * Unlike stack allocator, alignment value to be used for allocating object
 * must be specified when the instance is initialized, and cannot change
 * it after initialization.
 * The alignment argument of this function is the form of an amount of
 * bit shift.
 *
 * The max alignment value to be specified is the same as the alignment
 * value of the backend page allocator.  To get this value,
 * obs_pa_ops_ops_get_alignment() or obs_pa_curops_get_alignment() function
 * can be used.
 *
 *
 * @subsection slab_thread thread safety
 * ::OBS_slab structure cannot be accessed at the same time.
 * It is thread safe only if an ::OBS_slab is keeped exclusive among all
 * threads, or per-thread ::OBS_slab structures are used.
 *
 *
 * @section naming_convention naming convention
 *  - @em types :
 *      beginning OBS_, and the rests are lower case.
 *      Each words are separated by underscore.
 *      e.g. OBS_foo_bar
 *  - @em macros :
 *      beginning OBS_, and the rests are upper case.
 *      Each words are separated by underscore.
 *      e.g. OBS_FOO_BAR
 *  - @em functions :
 *      beginning obs_, and the rests are lower case.
 *      Each words are separated by underscore.
 *      e.g. obs_foo_bar
 *  - @em global @em variables :
 *      beginning obs, and the rests are mixed upper and lower case.
 *      Each words are separated by capitalization.
 *      e.g. obsFooBar.
 *
 * All symbols beginning OBS or obs are reserved, so users must not define such
 * symbols.  All symbols trailing underscore are private/protected, so users
 * must not touch them directly.
 *
 *
 * @section todo todo
 *  - page allocator for test.
 *  - add way for reserving many objects to slab allocator.
 *
 * @section author author
 * Takuya SHIOZAKI @<tshiozak at bsdclub dot org@>
 */

#include <obs/obs_stack.h>
#include <obs/obs_slab.h>

#endif
