/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs_pa_inspector.h>

#define DEFDEC(name)							      \
static void								      \
show_##name(OBS_pa_inspector_output_func func, void *closure,		      \
	   const OBS_pa_wrapper_event *ev, OBS_bool isarg)

void
show_region(
	OBS_pa_inspector_output_func func, void *closure,
	const OBS_region *reg)
{
	if (obs_region_is_valid(reg))
		(*func)(closure, "(%p,%u)", reg->head, (unsigned)reg->size);
	else
		(*func)(closure, "invalid");
}

DEFDEC(get_page_unit)
{
	if (!isarg)
		(*func)(closure, "get_page_unit=%u\n",
			(unsigned)*ev->result.get_page_unit.size);
}

DEFDEC(get_alignment)
{
	if (!isarg)
		(*func)(closure, "get_alignment=%d\n",
			(unsigned)*ev->result.get_alignment.alignment);
}

DEFDEC(get_available_features)
{
	if (!isarg)
		(*func)(closure, "get_available_features=0x%X\n",
			(unsigned)*ev->result.get_available_features.features);
}

DEFDEC(init)
{

	if (isarg)
		(*func)(closure, "init(num_units=%u, features=0x%X) = ",
			(unsigned)*ev->arg.init.num_units,
			(unsigned)*ev->arg.init.features);
	else
		(*func)(closure, "%s\n",
			obs_strstatus(*ev->result.init.status));
}

/*ARGSUSED*/
DEFDEC(uninit)
{
	if (isarg)
		(*func)(closure, "uninit\n");
}

static void
sized_op(OBS_pa_inspector_output_func func, void *closure,
	 const OBS_pa_wrapper_event *ev, OBS_bool isarg,
	 const char *name, OBS_bool isshowregarg)
{
	if (isarg) {
		(*func)(closure, "%s(", name);
		if (isshowregarg) {
			show_region(func, closure, ev->arg.alloc.region);
			(*func)(closure, ",%u) = ",
				(unsigned)*ev->arg.alloc.size);
		} else
			(*func)(closure, "%u) = ",
				(unsigned)*ev->arg.alloc.size);
	} else {
		(*func)(closure, "%s ",
			obs_strstatus(*ev->result.alloc.status));
		show_region(func, closure, ev->arg.alloc.region);
		(*func)(closure, "\n");
	}
}

DEFDEC(alloc)
{ sized_op(func, closure, ev, isarg, "alloc", OBS_FALSE); }

DEFDEC(expand_head)
{ sized_op(func, closure, ev, isarg, "expand_head", OBS_TRUE); }

DEFDEC(expand_tail)
{ sized_op(func, closure, ev, isarg, "expand_tail", OBS_TRUE); }

DEFDEC(shrink_head)
{ sized_op(func, closure, ev, isarg, "shrink_head", OBS_TRUE); }

DEFDEC(shrink_tail)
{ sized_op(func, closure, ev, isarg, "shrink_tail", OBS_TRUE); }

DEFDEC(concatenate)
{
	if (isarg) {
		(*func)(closure, "concatenate(");
		show_region(func, closure, ev->arg.concatenate.dstsrc1);
		(*func)(closure, ",");
		show_region(func, closure, ev->arg.concatenate.src2);
		(*func)(closure, ") = ");
	} else {
		(*func)(closure, "%s ",
			obs_strstatus(*ev->result.concatenate.status));
		show_region(func, closure, ev->arg.concatenate.dstsrc1);
		(*func)(closure, ",");
		show_region(func, closure, ev->arg.concatenate.src2);
		(*func)(closure, "\n");
	}
}

DEFDEC(free)
{
	if (isarg) {
		(*func)(closure, "free");
		show_region(func, closure, ev->arg.free.region);
		(*func)(closure, "\n");
	}
}

#define O_(t, f) OBS_ARRAY_INIT_(OBS_PA_WRAPPER_T_##t, &show_##f)

static void (*decoders[])(OBS_pa_inspector_output_func, void *,
			  const OBS_pa_wrapper_event *, OBS_bool) =
{
	O_(GET_PAGE_UNIT, get_page_unit),
	O_(GET_ALIGNMENT, get_alignment),
	O_(GET_AVAILABLE_FEATURES, get_available_features),
	O_(INIT, init),
	O_(UNINIT, uninit),
	O_(ALLOC, alloc),
	O_(EXPAND_HEAD, expand_head),
	O_(EXPAND_TAIL, expand_tail),
	O_(SHRINK_HEAD, shrink_head),
	O_(SHRINK_TAIL, shrink_tail),
	O_(CONCATENATE, concatenate),
	O_(FREE, free)
};


void
obs_pa_inspector_show_wrapper_event(OBS_pa_inspector_output_func func,
				    void *closure,
				    const OBS_pa_wrapper_event *ev)
{
	int type = OBS_PA_WRAPPER_TYPE(ev->type);

	if (type < 0 || type >= OBS_NUM_OF(decoders))
		(*func)(closure, "(unknown event)\n");
	else
		(*decoders[type])(func, closure, ev,
				  OBS_PA_WRAPPER_IS_PRE(ev->type));
}

static void
handler(void *closure, OBS_pa_wrapper_event *ev)
{
	OBS_pa_inspector *insp = closure;
	obs_pa_inspector_show_wrapper_event(
		insp->output_func, insp->closure, ev);
}

const OBS_pa_ops *
obs_inspect_pa_ops(
	OBS_pa_inspector *insp, const OBS_pa_ops *wrapped_ops,
	OBS_pa_inspector_output_func func, void *closure)
{
	insp->output_func = func;
	insp->closure = closure;
	return obs_wrap_pa_ops(&insp->wrapper, wrapped_ops, &handler, insp);
}
