/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs_pagealloc.h>

#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

OBS_PA_MAKE_OPS_(static, mm);

const OBS_pa_ops * const obsPaDefaultOps_ = &mm_ops;
const OBS_pa_ops *obsPaCurrentOps_ = &mm_ops;

OBS_size
/*ARGSUSED*/
mm_get_page_unit(const OBS_pa_ops *ops)
{
	static volatile OBS_size pagesize = 0;

	if (pagesize == 0)
		pagesize = sysconf(_SC_PAGESIZE);

	return pagesize;
}

int
/*ARGSUSED*/
mm_get_alignment(const OBS_pa_ops *ops)
{
	static volatile int align = 0;

	if (align == 0) {
		OBS_size pagesize = mm_get_page_unit(NULL);
		int i;
		for (i=0; pagesize; i++, pagesize >>= 1)
			if ((pagesize & 1))
				break;
		if (i < OBS_LEAST_MAX_ALIGNMENT)
			abort();
		align = i;
	}
	return align;
}

#define FEATURES ((OBS_flags)(OBS_PA_F_SHRINK_HEAD |			      \
			      OBS_PA_F_SHRINK_TAIL |			      \
			      OBS_PA_F_CONCATENATE))
OBS_flags
/*ARGSUSED*/
mm_get_available_features(const OBS_pa_ops *ops)
{
	return FEATURES;
}

OBS_status
mm_init(const OBS_pa_ops *ops, OBS_pa *pa, OBS_size num_units,
	OBS_flags features)
{
	OBS_size pagesize;

	if ((features & ~FEATURES) != 0)
		return OBS_S_NOTSUPP;

	pagesize = mm_get_page_unit(ops);
	if (num_units > OBS_SIZE_MAX/pagesize)
		return OBS_S_INVAL;

	memset(pa, 0, sizeof (*pa));

	pa->ops_ = ops;
	pa->priv_ = NULL;
	pa->block_size_ = pagesize * num_units;
	pa->features_ = features;

	return OBS_S_NOERR;
}

void mm_uninit(OBS_pa *pa)
{
	pa->ops_ = NULL;
}

OBS_status
mm_alloc(OBS_pa *pa, OBS_region *page, OBS_size num_pages)
{
	int saved;
	OBS_status ret;
	void *ptr;
	OBS_size size;

	if (num_pages > OBS_SIZE_MAX/pa->block_size_)
		return OBS_S_INVAL;

	saved = errno;
	size = pa->block_size_*num_pages;
	/* XXX: assumes MAP_ANON available. */
	ptr = mmap(NULL, size, PROT_READ|PROT_WRITE,
		   MAP_ANON|MAP_PRIVATE, -1, 0);
	if (ptr == MAP_FAILED) {
		if (errno == ENOMEM)
			ret = OBS_S_NOMEM;
		else
			ret = OBS_S_UNKNOWN;
		errno = saved;
		return ret;
	}

	page->head = ptr;
	page->size = size;

	return OBS_S_NOERR;
}

/*
 * normal mmap() mechanism lacks reliable expansile method.
 *   - MAP_FIXED cannot use for such purpose because
 *     it may replace old mapping and there is no reliable and
 *     thread-safe method to know whether a page is mapped.
 *   - posix_typed_mem_open() does not exist on many platforms,
 *     and it is not handy.
 *   - MAP_TRYFIXED does unnecessary fallback,
 *     and it is not standard.
 */
OBS_status
/*ARGSUSED*/
mm_expand_head(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{ return OBS_S_NOTSUPP; }

OBS_status
/*ARGSUSED*/
mm_expand_tail(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{ return OBS_S_NOTSUPP; }

OBS_status
mm_shrink_head(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{

	if ((pa->features_ & OBS_PA_F_SHRINK_HEAD) == 0)
		return OBS_S_NOTSUPP;

	if (decr > OBS_SIZE_MAX/pa->block_size_)
		return OBS_S_INVAL; /* overflow */

	if (pages->head == NULL)
		return OBS_S_NOERR;

	decr *= pa->block_size_;

	if (pages->size < decr)
		decr = pages->size;

	if (munmap(pages->head, decr))
		return OBS_S_INVAL;

	pages->head = (void *)((char *)pages->head + decr);
	pages->size -= decr;
	if (pages->size == 0)
		pages->head = NULL;

	return OBS_S_NOERR;
}

OBS_status
mm_shrink_tail(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{

	if ((pa->features_ & OBS_PA_F_SHRINK_TAIL) == 0)
		return OBS_S_NOTSUPP;

	if (decr > OBS_SIZE_MAX/pa->block_size_)
		return OBS_S_INVAL; /* overflow */

	if (pages->head == NULL)
		return OBS_S_NOERR;

	decr *= pa->block_size_;

	if (pages->size < decr)
		decr = pages->size;

	if (munmap((char *)pages->head + pages->size - decr, decr))
		return OBS_S_INVAL;

	pages->size -= decr;
	if (pages->size == 0)
		pages->head = NULL;

	return OBS_S_NOERR;
}

OBS_status
/*ARGSUSED*/
mm_concatenate(OBS_pa *pa, OBS_region *dstsrc1, OBS_region *src2)
{

	if ((pa->features_ & OBS_PA_F_CONCATENATE) == 0)
		return OBS_S_NOTSUPP;

	if (!obs_region_is_contiguous(dstsrc1, src2))
		return OBS_S_NOT_CONTIGUOUS;

	dstsrc1->size += src2->size;
	src2->head = NULL;
	src2->size = 0;

	return OBS_S_NOERR;
}

void
/*ARGSUSED*/
mm_free(OBS_pa *pa, OBS_region *pages)
{
	if (pages->head)
		munmap(pages->head, pages->size);
	pages->head = NULL;
	pages->size = 0;
}
