/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs_pagealloc.h>

#include <Windows.h>

OBS_PA_MAKE_OPS_(static, winva);

const OBS_pa_ops * const obsPaDefaultOps_ = &winva_ops;
const OBS_pa_ops *obsPaCurrentOps_ = &winva_ops;

OBS_size
/*ARGSUSED*/
winva_get_page_unit(const OBS_pa_ops *ops)
{
	static volatile OBS_size pagesize = 0;

	if (pagesize == 0) {
		SYSTEM_INFO si;

		GetSystemInfo(&si);
		pagesize = si.dwPageSize;
	}

	return pagesize;
}

int
/*ARGSUSED*/
winva_get_alignment(const OBS_pa_ops *ops)
{
	static volatile int align = 0;

	if (align == 0) {
		OBS_size pagesize = winva_get_page_unit(NULL);
		int i;
		for (i=0; pagesize; i++, pagesize >>= 1)
			if ((pagesize & 1))
				break;
		if (i < OBS_LEAST_MAX_ALIGNMENT)
			abort();
		align = i;
	}
	return align;
}

#define FEATURES ((OBS_flags)(0))
OBS_flags
/*ARGSUSED*/
winva_get_available_features(const OBS_pa_ops *ops)
{
	return FEATURES;
}

OBS_status
winva_init(const OBS_pa_ops *ops, OBS_pa *pa, OBS_size num_units,
	OBS_flags features)
{
	OBS_size pagesize;

	if ((features & ~FEATURES) != 0)
		return OBS_S_NOTSUPP;

	pagesize = winva_get_page_unit(ops);
	if (num_units > OBS_SIZE_MAX/pagesize)
		return OBS_S_INVAL;

	memset(pa, 0, sizeof (*pa));

	pa->ops_ = ops;
	pa->priv_ = NULL;
	pa->block_size_ = pagesize * num_units;
	pa->features_ = features;

	return OBS_S_NOERR;
}

void winva_uninit(OBS_pa *pa)
{
	pa->ops_ = NULL;
}

OBS_status
winva_alloc(OBS_pa *pa, OBS_region *page, OBS_size num_pages)
{
	DWORD saved;
	OBS_status ret;
	void *ptr;
	OBS_size size;

	if (num_pages > OBS_SIZE_MAX/pa->block_size_)
		return OBS_S_INVAL;

	saved = GetLastError();
	size = pa->block_size_*num_pages;
	ptr = VirtualAlloc(NULL, size, MEM_COMMIT, PAGE_READWRITE);
	if (ptr == NULL) {
		if (GetLastError() == ERROR_NOT_ENOUGH_MEMORY)
			ret = OBS_S_NOMEM;
		else
			ret = OBS_S_UNKNOWN;
		SetLastError(saved);
		return ret;
	}
	SetLastError(saved);

	page->head = ptr;
	page->size = size;

	return OBS_S_NOERR;
}

OBS_status
/*ARGSUSED*/
winva_expand_head(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{ return OBS_S_NOTSUPP; }

OBS_status
/*ARGSUSED*/
winva_expand_tail(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{ return OBS_S_NOTSUPP; }

OBS_status
/*ARGSUSED*/
winva_shrink_head(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{ return OBS_S_NOTSUPP; }

OBS_status
/*ARGSUSED*/
winva_shrink_tail(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{ return OBS_S_NOTSUPP; }

OBS_status
/*ARGSUSED*/
winva_concatenate(OBS_pa *pa, OBS_region *dstsrc1, OBS_region *src2)
{ return OBS_S_NOTSUPP; }

void
/*ARGSUSED*/
winva_free(OBS_pa *pa, OBS_region *pages)
{
	if (pages->head)
		VirtualFree(pages->head, 0, MEM_RELEASE);
	pages->head = NULL;
	pages->size = 0;
}
