/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs_pa_wrapper.h>
#include <string.h>

OBS_PA_MAKE_OPS_(static, wrapper);

OBS_DEF_INLINE_FUNC_(const OBS_pa_wrapper *, ops_to_wrapper,
		     (const OBS_pa_ops *ops))
{
	return (const void *)(
		(const char *)(const void *)ops -
		offsetof(OBS_pa_wrapper, ops));
}

OBS_DEF_INLINE_FUNC_(const OBS_pa_ops *, wrapper_to_ops,
		     (const OBS_pa_wrapper *pw))
{ return &pw->ops; }

OBS_DEF_INLINE_FUNC_(const OBS_pa_wrapper *, unwrap, (OBS_pa *pa))
{
	const OBS_pa_wrapper *pw = ops_to_wrapper(obs_pa_get_ops(pa));
	pa->ops_ = pw->wrapped_ops;
	return pw;
}

OBS_DEF_INLINE_FUNC_(void, wrap, (OBS_pa *pa, const OBS_pa_wrapper *pw))
{
	pa->ops_ = wrapper_to_ops(pw);
}

static OBS_size
wrapper_get_page_unit(const OBS_pa_ops *ops)
{
	OBS_size ret;
	const OBS_pa_wrapper *pw = ops_to_wrapper(ops);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_GET_PAGE_UNIT;
	ev.arg.get_page_unit.ops = &ops;
	(*pw->handler)(pw->closure, &ev);
	ret = obs_pa_ops_get_page_unit(pw->wrapped_ops);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.get_page_unit.size = &ret;
	(*pw->handler)(pw->closure, &ev);
	return ret;
}

static int
wrapper_get_alignment(const OBS_pa_ops *ops)
{
	int ret;
	const OBS_pa_wrapper *pw = ops_to_wrapper(ops);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_GET_ALIGNMENT;
	ev.arg.get_alignment.ops = &ops;
	(*pw->handler)(pw->closure, &ev);
	ret = obs_pa_ops_get_alignment(pw->wrapped_ops);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.get_alignment.alignment = &ret;
	(*pw->handler)(pw->closure, &ev);
	return ret;
}

static OBS_flags
wrapper_get_available_features(const OBS_pa_ops *ops)
{
	OBS_flags ret;
	const OBS_pa_wrapper *pw = ops_to_wrapper(ops);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_GET_AVAILABLE_FEATURES;
	ev.arg.get_available_features.ops = &ops;
	(*pw->handler)(pw->closure, &ev);
	ret = obs_pa_ops_get_available_features(pw->wrapped_ops);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.get_available_features.features = &ret;
	(*pw->handler)(pw->closure, &ev);

	return ret;
}

static OBS_status
wrapper_init(const OBS_pa_ops *ops, OBS_pa *pa, OBS_size num_units,
	     OBS_flags features)
{
	OBS_status ret;
	const OBS_pa_wrapper *pw = ops_to_wrapper(ops);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_INIT;
	ev.arg.init.ops = &ops;
	ev.arg.init.pa = pa;
	ev.arg.init.num_units = &num_units;
	ev.arg.init.features = &features;
	(*pw->handler)(pw->closure, &ev);
	ret = obs_pa_ops_init(pw->wrapped_ops, pa, num_units, features);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.init.status = &ret;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);

	return ret;
}

static void
wrapper_uninit(OBS_pa *pa)
{
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_UNINIT;
	ev.arg.uninit.pa = pa;
	(*pw->handler)(pw->closure, &ev);
	(*pw->wrapped_ops->uninit_)(pa);
	ev.type |= OBS_PA_WRAPPER_POST;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);
}

static OBS_status
wrapper_alloc(OBS_pa *pa, OBS_region *pages, OBS_size num_pages)
{
	OBS_status ret;
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_ALLOC;
	ev.arg.alloc.pa = pa;
	ev.arg.alloc.region = pages;
	ev.arg.alloc.size = &num_pages;
	(*pw->handler)(pw->closure, &ev);
	ret = (*pw->wrapped_ops->alloc_)(pa, pages, num_pages);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.alloc.status = &ret;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);

	return ret;
}

static OBS_status
wrapper_expand_head(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{
	OBS_status ret;
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_EXPAND_HEAD;
	ev.arg.expand_head.pa = pa;
	ev.arg.expand_head.region = pages;
	ev.arg.expand_head.size = &incr;
	(*pw->handler)(pw->closure, &ev);
	ret = (*pw->wrapped_ops->expand_head_)(pa, pages, incr);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.expand_head.status = &ret;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);

	return ret;
}

static OBS_status
wrapper_expand_tail(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{
	OBS_status ret;
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_EXPAND_TAIL;
	ev.arg.expand_tail.pa = pa;
	ev.arg.expand_tail.region = pages;
	ev.arg.expand_tail.size = &incr;
	(*pw->handler)(pw->closure, &ev);
	ret = (*pw->wrapped_ops->expand_tail_)(pa, pages, incr);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.expand_tail.status = &ret;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);

	return ret;
}

static OBS_status
wrapper_shrink_head(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{
	OBS_status ret;
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_SHRINK_HEAD;
	ev.arg.shrink_head.pa = pa;
	ev.arg.shrink_head.region = pages;
	ev.arg.shrink_head.size = &decr;
	(*pw->handler)(pw->closure, &ev);
	ret = (*pw->wrapped_ops->shrink_head_)(pa, pages, decr);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.shrink_head.status = &ret;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);

	return ret;
}

static OBS_status
wrapper_shrink_tail(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{
	OBS_status ret;
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_SHRINK_TAIL;
	ev.arg.shrink_tail.pa = pa;
	ev.arg.shrink_tail.region = pages;
	ev.arg.shrink_tail.size = &decr;
	(*pw->handler)(pw->closure, &ev);
	ret = (*pw->wrapped_ops->shrink_tail_)(pa, pages, decr);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.shrink_tail.status = &ret;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);

	return ret;
}

static OBS_status
wrapper_concatenate(OBS_pa *pa, OBS_region *dstsrc1, OBS_region *src2)
{
	OBS_status ret;
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_CONCATENATE;
	ev.arg.concatenate.pa = pa;
	ev.arg.concatenate.dstsrc1 = dstsrc1;
	ev.arg.concatenate.src2 = src2;
	(*pw->handler)(pw->closure, &ev);
	ret = (*pw->wrapped_ops->concatenate_)(pa, dstsrc1, src2);
	ev.type |= OBS_PA_WRAPPER_POST;
	ev.result.concatenate.status = &ret;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);

	return ret;
}

static void
wrapper_free(OBS_pa *pa, OBS_region *pages)
{
	const OBS_pa_wrapper *pw = unwrap(pa);
	OBS_pa_wrapper_event ev;

	ev.type = OBS_PA_WRAPPER_T_FREE;
	ev.arg.free.pa = pa;
	ev.arg.free.region = pages;
	(*pw->handler)(pw->closure, &ev);
	(*pw->wrapped_ops->free_)(pa, pages);
	ev.type |= OBS_PA_WRAPPER_POST;
	(*pw->handler)(pw->closure, &ev);
	wrap(pa, pw);
}


const OBS_pa_ops *
obs_wrap_pa_ops(OBS_pa_wrapper *pw,
		const OBS_pa_ops *wrapped_ops,
		OBS_pa_wrapper_event_handler handler,
		void *closure)
{
	pw->wrapped_ops = wrapped_ops;
	pw->handler = handler;
	pw->closure = closure;
	memcpy(&pw->ops, &wrapper_ops, sizeof (OBS_pa_ops));
	return wrapper_to_ops(pw);
}
