/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file */

#ifndef OBS_PA_WRAPPED_H_
#define OBS_PA_WRAPPED_H_

#include <obs/obs_pagealloc.h>

/** wrapped page allocator class object. */
typedef struct OBS_pa_wrapper_ops OBS_pa_wrapper_ops;

/** mask of an bit indicating whether the event is pre-call of the operator. */
#define OBS_PA_WRAPPER_PRE_POST_MASK	0x10000000
/** indicating the event is pre-calling hook of the operator. */
#define OBS_PA_WRAPPER_PRE		0x00000000
/** indicating the event is post-calling hook of the operator. */
#define OBS_PA_WRAPPER_POST		0x10000000
/** mask for the kind of operators. */
#define OBS_PA_WRAPPER_TYPE_MASK	0x0FFFFFFF
/** method type code indicating the event hooks get_page_unit method. */
#define OBS_PA_WRAPPER_T_GET_PAGE_UNIT		0
/** method type code indicating the event hooks get_alignment method. */
#define OBS_PA_WRAPPER_T_GET_ALIGNMENT		1
/** method type code indicating the event hooks get_available_features
 * method. */
#define OBS_PA_WRAPPER_T_GET_AVAILABLE_FEATURES	2
/** method type code indicating the event hooks init method. */
#define OBS_PA_WRAPPER_T_INIT			3
/** method type code indicating the event hooks uninit method. */
#define OBS_PA_WRAPPER_T_UNINIT			4
/** method type code indicating the event hooks alloc method. */
#define OBS_PA_WRAPPER_T_ALLOC			5
/** method type code indicating the event hooks expand_head method. */
#define OBS_PA_WRAPPER_T_EXPAND_HEAD		6
/** method type code indicating the event hooks expand_tail method. */
#define OBS_PA_WRAPPER_T_EXPAND_TAIL		7
/** method type code indicating the event hooks shrink_head method. */
#define OBS_PA_WRAPPER_T_SHRINK_HEAD		8
/** method type code indicating the event hooks shrink_tail method. */
#define OBS_PA_WRAPPER_T_SHRINK_TAIL		9
/** method type code indicating the event hooks concatinate method. */
#define OBS_PA_WRAPPER_T_CONCATENATE		10
/** method type code indicating the event hooks free method. */
#define OBS_PA_WRAPPER_T_FREE			11
/** check whether the type is indicating pre-calling. */
#define OBS_PA_WRAPPER_IS_PRE(t)					      \
(((t) & OBS_PA_WRAPPER_PRE_POST_MASK) == OBS_PA_WRAPPER_PRE)
/** check whether the type is indicating post-calling. */
#define OBS_PA_WRAPPER_IS_POST(t)					      \
(((t) & OBS_PA_WRAPPER_PRE_POST_MASK) == OBS_PA_WRAPPER_POST)
/** get method type code. */
#define OBS_PA_WRAPPER_TYPE(t)						      \
((t) & OBS_PA_WRAPPER_TYPE_MASK)

/** simple argument containing an pointer to OBS_pa_ops. */
typedef struct OBS_pa_wrapper_ops_arg
{
	/** pointer of pointer to the OBS_pa_ops. */
	const OBS_pa_ops **ops;
} OBS_pa_wrapper_ops_arg;

/** arguments for init method. */
typedef struct OBS_pa_wrapper_init_arg
{
	/** pointer of pointer to the OBS_pa_ops. */
	const OBS_pa_ops **ops;
	/** pointer of OBS_pa. */
	OBS_pa *pa;
	/** pointer to num_units. */
	OBS_size *num_units;
	/** pointer to features. */
	OBS_flags *features;
} OBS_pa_wrapper_init_arg;

/** simple argument containing an pointer to OBS_pa. */
typedef struct OBS_pa_wrapper_pa_arg
{
	/** pointer to OBS_pa. */
	OBS_pa *pa;
} OBS_pa_wrapper_pa_arg;

/** arguments for alloc / expand_* / shrink_* methods. */
typedef struct OBS_pa_wrapper_regsize_arg
{
	/** pointer to OBS_pa. */
	OBS_pa *pa;
	/** pointer to region. */
	OBS_region *region;
	/** pointer to size. */
	OBS_size *size;
} OBS_pa_wrapper_regsize_arg;

/** arguments for concatinate method. */
typedef struct OBS_pa_wrapper_concat_arg
{
	/** pointer to OBS_pa. */
	OBS_pa *pa;
	/** pointer to dstsrc1. */
	OBS_region *dstsrc1;
	/** pointer to src2. */
	OBS_region *src2;
} OBS_pa_wrapper_concat_arg;

/** arguments for free method */
typedef struct OBS_pa_wrapper_free_arg
{
	/** pointer to OBS_pa. */
	OBS_pa *pa;
	/** pointer to region. */
	OBS_region *region;
} OBS_pa_wrapper_free_arg;

/** result returning a size. */
typedef struct OBS_pa_wrapper_size_res
{
	/** pointer to size. */
	OBS_size *size;
} OBS_pa_wrapper_size_res;

/** result returning a shift amount of an alignment. */
typedef struct OBS_pa_wrapper_align_res
{
	/** pointer to alignment. */
	int *alignment;
} OBS_pa_wrapper_align_res;

/** result returning a feature mask. */
typedef struct OBS_pa_wrapper_feat_res
{
	/** pointer to features. */
	OBS_flags *features;
} OBS_pa_wrapper_feat_res;

/** result returning a status. */
typedef struct OBS_pa_wrapper_status_res
{
	/** pointer to status. */
	OBS_status *status;
} OBS_pa_wrapper_status_res;

/** data of hooked method. */
typedef struct OBS_pa_wrapper_event
{
	/** event type. */
	int type;
	/** arguments. */
	union
	{
		OBS_pa_wrapper_ops_arg get_page_unit;
		OBS_pa_wrapper_ops_arg get_alignment;
		OBS_pa_wrapper_ops_arg get_available_features;
		OBS_pa_wrapper_init_arg init;
		OBS_pa_wrapper_pa_arg uninit;
		OBS_pa_wrapper_regsize_arg alloc;
		OBS_pa_wrapper_regsize_arg expand_head;
		OBS_pa_wrapper_regsize_arg expand_tail;
		OBS_pa_wrapper_regsize_arg shrink_head;
		OBS_pa_wrapper_regsize_arg shrink_tail;
		OBS_pa_wrapper_concat_arg concatenate;
		OBS_pa_wrapper_free_arg free;
	} arg;
	/** results. */
	union
	{
		OBS_pa_wrapper_size_res get_page_unit;
		OBS_pa_wrapper_align_res get_alignment;
		OBS_pa_wrapper_feat_res get_available_features;
		OBS_pa_wrapper_status_res init;
		OBS_pa_wrapper_status_res alloc;
		OBS_pa_wrapper_status_res expand_head;
		OBS_pa_wrapper_status_res expand_tail;
		OBS_pa_wrapper_status_res shrink_head;
		OBS_pa_wrapper_status_res shrink_tail;
		OBS_pa_wrapper_status_res concatenate;
	} result;
} OBS_pa_wrapper_event;

/** pointer type to the wrapper function. */
typedef void (*OBS_pa_wrapper_event_handler)(void *, OBS_pa_wrapper_event *);

/** wrapper allocator class object. */
typedef struct OBS_pa_wrapper
{
	/**@internal page allocator class object to be wrapped. */
	const OBS_pa_ops *wrapped_ops;
	/**@internal wrapper function. */
	OBS_pa_wrapper_event_handler handler;
	/**@internal user data of the handler. */
	void *closure;
	/**@internal vector of method for this allocator. */
	OBS_pa_ops ops;
} OBS_pa_wrapper;

OBS_BEGIN_DECL_
/**
 * initialize a OBS_pa_wrapper class object.
 *
 * @param newops	OBS_pa_wrapper object to be initialized.
 * @param wrapped_ops	allocator class object to be wrapped.
 * @param handler	wrapper function.
 * @param closure	userdata of the handler.
 * @return		base class object of @a newops.
 */
const OBS_pa_ops *obs_wrap_pa_ops(OBS_pa_wrapper *newops,
				  const OBS_pa_ops *wrapped_ops,
				  OBS_pa_wrapper_event_handler handler,
				  void *closure);
OBS_END_DECL_

#endif
