/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file
 * page allocator.
 */

#ifndef OBS_PAGEALLOC_H_
#define OBS_PAGEALLOC_H_

#include <obs/obs_config.h>
#include <obs/obs_types.h>
#include <obs/obs_defines.h>
#include <obs/obs_status.h>
#include <obs/obs_region.h>

/**@name PageAllocFeatures
 * features of page allocator.
 * @{
 */
/** feature flag indicating that expand tail is enabled. */
#define OBS_PA_F_EXPAND_TAIL		0x0001U
/** feature flag indicating that expand head is enabled. */
#define OBS_PA_F_EXPAND_HEAD		0x0002U
/** feature flag indicating that shrink tail is enabled. */
#define OBS_PA_F_SHRINK_TAIL		0x0004U
/** feature flag indicating that shrink head is enabled. */
#define OBS_PA_F_SHRINK_HEAD		0x0008U
/** feature flag indicating that all free on uninit is enabled. */
#define OBS_PA_F_ALL_FREE_ON_UNINIT	0x0010U
/** feature flag indicating that concatenate is enabled. */
#define OBS_PA_F_CONCATENATE		0x0020U

/** check a page allocator whether a feature is enabled. */
#define OBS_PA_IS_F(pa, tag)						      \
	(((pa)->features_ & OBS_PA_F_##tag) != 0)
/** check feature mask value whether a feature is enabled. */
#define OBS_PA_FEATURES_IS_F(features, tag)				      \
	(((features) & OBS_PA_F_##tag) != 0)

/**@}*/


/** the allocator-class-specific data per allocator. */
typedef struct OBS_pa_private OBS_pa_private;
/** a page allocator. */
typedef struct OBS_pa OBS_pa;
/** a class object for page allocator type. */
typedef struct OBS_pa_ops OBS_pa_ops;

/** base portion of the class object. */
struct OBS_pa_ops
{
	/**@internal get page unit size. */
	OBS_size (*get_page_unit_)(const OBS_pa_ops *);
	/**@internal get alignment in a bytes. */
	int (*get_alignment_)(const OBS_pa_ops *);
	/**@internal get mask of available features. */
	OBS_flags (*get_available_features_)(const OBS_pa_ops *);
	/**@internal initialize an allocator instance. */
	OBS_status (*init_)(const OBS_pa_ops *, OBS_pa *, OBS_size, OBS_flags);
	/**@internal uninitialize an allocator instance. */
	void (*uninit_)(OBS_pa *);
	/**@internal allocate new blocks. */
	OBS_status (*alloc_)(OBS_pa *, OBS_region *, OBS_size);
	/**@internal expand toward head. */
	OBS_status (*expand_head_)(OBS_pa *, OBS_region *, OBS_size);
	/**@internal expand toward tail. */
	OBS_status (*expand_tail_)(OBS_pa *, OBS_region *, OBS_size);
	/**@internal shrink head side. */
	OBS_status (*shrink_head_)(OBS_pa *, OBS_region *, OBS_size);
	/**@internal shrink tail side. */
	OBS_status (*shrink_tail_)(OBS_pa *, OBS_region *, OBS_size);
	/**@internal concatenate two regions. */
	OBS_status (*concatenate_)(OBS_pa *, OBS_region *, OBS_region *);
	/**@internal free blocks. */
	void (*free_)(OBS_pa *, OBS_region *);
};

/** make base portion of OBS_pa_ops. */
#define OBS_PA_MAKE_OPS_(scope, name)					      \
static OBS_size name##_get_page_unit(const OBS_pa_ops *);		      \
static int name##_get_alignment(const OBS_pa_ops *);			      \
static OBS_flags name##_get_available_features(const OBS_pa_ops *);	      \
static OBS_status name##_init(						      \
	const OBS_pa_ops *, OBS_pa *, OBS_size, OBS_flags);		      \
static void name##_uninit(OBS_pa *);					      \
static OBS_status name##_alloc(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_expand_head(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_expand_tail(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_shrink_head(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_shrink_tail(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_concatenate(OBS_pa *, OBS_region *, OBS_region *);   \
static void name##_free(OBS_pa *, OBS_region *);			      \
scope const OBS_pa_ops name##_ops =					      \
{									      \
	OBS_STRUCT_INIT_(get_page_unit_, name##_get_page_unit),		      \
	OBS_STRUCT_INIT_(get_alignment_, name##_get_alignment),		      \
	OBS_STRUCT_INIT_(get_available_features_,			      \
			 name##_get_available_features),		      \
	OBS_STRUCT_INIT_(init_, name##_init),				      \
	OBS_STRUCT_INIT_(uninit_, name##_uninit),			      \
	OBS_STRUCT_INIT_(alloc_, name##_alloc),				      \
	OBS_STRUCT_INIT_(expand_head_, name##_expand_head),		      \
	OBS_STRUCT_INIT_(expand_tail_, name##_expand_tail),		      \
	OBS_STRUCT_INIT_(shrink_head_, name##_shrink_head),		      \
	OBS_STRUCT_INIT_(shrink_tail_, name##_shrink_tail),		      \
	OBS_STRUCT_INIT_(concatenate_, name##_concatenate),		      \
	OBS_STRUCT_INIT_(free_, name##_free)				      \
}

/** page allocator instance (opaque). */
struct OBS_pa
{
	/** @internal class object. */
	const OBS_pa_ops *ops_;
	/** @internal class-specific data. */
	OBS_pa_private *priv_;
	/** @internal block unit size. */
	OBS_size block_size_;
	/** @internal feature mask */
	OBS_flags features_;
};

/** @internal pointer to the default page allocator class object. */
extern const OBS_pa_ops * const obsPaDefaultOps_;
/** @internal pointer to the current page allocator class object. */
extern const OBS_pa_ops *obsPaCurrentOps_;

/**
 * set current allocator class for the allocator instance initialization.
 *
 * @param ops	page allocator class object to be set.
 */
OBS_DEF_INLINE_FUNC_(void, obs_pa_set_current_ops, (const OBS_pa_ops *ops))
{ obsPaCurrentOps_ = ops; }

/**
 * get current allocator class for the allocator instance initialization.
 *
 * @return	current page allocator class object.
 */
OBS_DEF_INLINE_FUNC_(const OBS_pa_ops *, obs_pa_get_current_ops, (void))
{ return obsPaCurrentOps_; }

/**
 * get default allocator class under the platform.
 *
 * @return	default page allocator class object.
 */
OBS_DEF_INLINE_FUNC_(const OBS_pa_ops *, obs_pa_get_default_ops, (void))
{ return obsPaDefaultOps_; }

/**
 * get page unit size on the specified allocator class.
 *
 * @param ops	page allocator class object.
 * @return	page unit size in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_pa_ops_get_page_unit,
		     (const OBS_pa_ops *ops))
{ return (*ops->get_page_unit_)(ops); }

/**
 * get page unit size on the current allocator class.
 *
 * @return	page unit size in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_pa_curops_get_page_unit, (void))
{ return obs_pa_ops_get_page_unit(obsPaCurrentOps_); }

/**
 * get mask indicating the available features on the specified allocator class.
 *
 * @param ops	page allocator class object.
 * @return	feature mask, OR-ed value of OBS_PA_F_* macros.
 */
OBS_DEF_INLINE_FUNC_(OBS_flags, obs_pa_ops_get_available_features,
		     (const OBS_pa_ops *ops))
{ return (*ops->get_available_features_)(ops); }

/**
 * get mask indicating the available features on the current allocator class.
 *
 * @return	feature mask, OR-ed value of OBS_PA_F_* macros.
 */
OBS_DEF_INLINE_FUNC_(OBS_flags, obs_pa_curops_get_available_features, (void))
{ return obs_pa_ops_get_available_features(obsPaCurrentOps_); }

/**
 * get shift amount of alignment value on the specified allocator class.
 *
 * @param ops	page allocator class object.
 * @return	shift amount of alignment value.
 * @note	All page allocators ensures this value equal to or more than 8.
 *		Typically, this value is the same as the shift amount of the
 *		page size.
 */
OBS_DEF_INLINE_FUNC_(int, obs_pa_ops_get_alignment,
		     (const OBS_pa_ops *ops))
{ return (*ops->get_alignment_)(ops); }

/**
 * get alignment value on the current allocator class.
 *
 * @return	alignment value in a byte.
 */
OBS_DEF_INLINE_FUNC_(int, obs_pa_curops_get_alignment, (void))
{ return obs_pa_ops_get_alignment(obsPaCurrentOps_); }

/**
 * initialize an allocator instance in the specified allocator class.
 *
 * @param ops		page allocator class object.
 * @param pa		page allocator instance object to be initialized.
 * @param num_units	block size in the page unit.
 * @param features	feature mask.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_ops_init,
		     (const OBS_pa_ops *ops, OBS_pa *pa,
		      OBS_size num_units, OBS_flags features))
{ return (*ops->init_)(ops, pa, num_units, features); }

/**
 * initialize an allocator instance in the default allocator class.
 *
 * @param pa		page allocator instance object to be initialized.
 * @param num_units	block size in the page unit.
 * @param features	feature mask.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_curops_init,
		     (OBS_pa *pa, OBS_size num_units, OBS_flags features))
{ return obs_pa_ops_init(obsPaCurrentOps_, pa, num_units, features); }

/**
 * uninitialize an allocator.
 *
 * @param pa		page allocator instance object to be uninitialized.
 */
OBS_DEF_INLINE_FUNC_(void, obs_pa_uninit, (OBS_pa *pa))
{ (*pa->ops_->uninit_)(pa); }

/**
 * allocate new blocks.
 *
 * @param pa		page allocator instance.
 * @param blocks	pointer to the region for storing the result.
 * @param num_blocks	number of blocks to be allocated.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_alloc,
		     (OBS_pa *pa, OBS_region *blocks, OBS_size num_blocks))
{ return (*pa->ops_->alloc_)(pa, blocks, num_blocks); }

/**
 * expand toward head side.
 *
 * @param pa		page allocator instance.
 * @param blocks	pointer to the region to be expanded.
 * @param incr		increment of blocks.
 * @return		result status.
 * @note this function is available if OBS_PA_F_EXPAND_HEAD is specified.
 *       Otherwise, this will cause OBS_S_NOTSUPP error.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_expand_head,
		     (OBS_pa *pa, OBS_region *blocks, OBS_size incr))
{ return (*pa->ops_->expand_head_)(pa, blocks, incr); }

/**
 * expand toward tail side.
 *
 * @param pa		page allocator instance.
 * @param blocks	pointer to the region to be expanded.
 * @param incr		increment of blocks.
 * @return		result status.
 * @note this function is available if OBS_PA_F_EXPAND_TAIL is specified.
 *       Otherwise, this will cause OBS_S_NOTSUPP error.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_expand_tail,
		     (OBS_pa *pa, OBS_region *blocks, OBS_size incr))
{ return (*pa->ops_->expand_tail_)(pa, blocks, incr); }

/**
 * shrink head side of the specified blocks.
 *
 * @param pa		page allocator instance.
 * @param blocks	pointer to the region to be expanded.
 * @param decr		decrement of blocks.
 * @return		result status.
 * @note this function is available if OBS_PA_F_SHRINK_HEAD is specified.
 *       Otherwise, this will cause OBS_S_NOTSUPP error.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_shrink_head,
		     (OBS_pa *pa, OBS_region *blocks, OBS_size decr))
{ return (*pa->ops_->shrink_head_)(pa, blocks, decr); }

/**
 * shrink tail side of the specified blocks.
 *
 * @param pa		page allocator instance.
 * @param blocks	pointer to the region to be expanded.
 * @param decr		decrement of blocks.
 * @return		result status.
 * @note this function is available if OBS_PA_F_SHRINK_TAIL is specified.
 *       Otherwise, this will cause OBS_S_NOTSUPP error.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_shrink_tail,
		     (OBS_pa *pa, OBS_region *blocks, OBS_size decr))
{ return (*pa->ops_->shrink_tail_)(pa, blocks, decr); }

/**
 * free blocks.
 *
 * @param pa		page allocator instance.
 * @param blocks	pointer to the region to be expanded.
 */
OBS_DEF_INLINE_FUNC_(void, obs_pa_free, (OBS_pa *pa, OBS_region *blocks))
{ (*pa->ops_->free_)(pa, blocks); }

/**
 * concatenate contiguous two blocks.
 *
 * @param pa		page allocator instance.
 * @param dstsrc1	pointer to the region which has lower address.
 * @param src2		pointer to the region which has higher address.
 * @note this function is available if OBS_PA_F_CONCATENATE is specified.
 *       Otherwise, this will cause OBS_S_NOTSUPP error.
 *       src2 will invalidated after success.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_concatenate,
		     (OBS_pa *pa, OBS_region *dstsrc1, OBS_region *src2))
{ return (*pa->ops_->concatenate_)(pa, dstsrc1, src2); }

/**
 * get block size of the allocator instance.
 *
 * @param pa	page allocator instance.
 * @return	block size in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_pa_get_block_size, (const OBS_pa *pa))
{ return pa->block_size_; }

/**
 * get current specified features for the allocator instance.
 *
 * @param pa	page allocator instance.
 * @return	feature mask.
 */
OBS_DEF_INLINE_FUNC_(OBS_flags, obs_pa_get_features, (const OBS_pa *pa))
{ return pa->features_; }

/**
 * get class object of the allocator instance.
 *
 * @param pa	page allocator instance.
 * @return	allocator class object.
 */
OBS_DEF_INLINE_FUNC_(const OBS_pa_ops *, obs_pa_get_ops, (const OBS_pa *pa))
{ return pa->ops_; }

#endif
