/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file*/

#ifndef OBS_REGION_H_
#define OBS_REGION_H_

#include <obs/obs_config.h>
#include <obs/obs_types.h>
#include <obs/obs_defines.h>

/** region (transparent structure) */
typedef struct OBS_region OBS_region;

/** region (transparent structure) */
struct OBS_region
{
	/** head of the region. */
	void *head;
	/** region size in a byte. */
	OBS_size size;
};

/**
 * initialize a region.
 *
 * @param reg		region.
 */
OBS_DEF_INLINE_FUNC_(void, obs_region_init, (OBS_region *reg))
{
	reg->head = NULL;
	reg->size = 0;
}

/**
 * check whether a region is valid.
 *
 * @param reg		region
 * @return		true if the region is valid.
 */
OBS_DEF_INLINE_FUNC_(OBS_bool, obs_region_is_valid, (const OBS_region *reg))
{ return reg->head != NULL; }

/**
 * check whether a region is valid and zero-sized region.
 *
 * @param reg		region
 * @return		true if the region is valid and its size is zero.
 */
OBS_DEF_INLINE_FUNC_(OBS_bool, obs_region_is_zero, (const OBS_region *reg))
{ return reg->head && reg->size == 0; }

/**
 * check whether two regions are contiguous.
 *
 * @param reg1		lower region.
 * @param reg2		higher region.
 * @return		true if reg2 is contiguous just after reg1.
 */
OBS_DEF_INLINE_FUNC_(OBS_bool, obs_region_is_contiguous,
		     (const OBS_region *reg1, const OBS_region *reg2))
{
	return reg1->head && reg2->head &&
	    (const char *)reg1->head + reg1->size == reg2->head;
}

/**
 * check whether a pointer is included in a region.
 *
 * @param reg		region.
 * @param ptr		pointer.
 * @return		true if ptr is included in reg.
 */
OBS_DEF_INLINE_FUNC_(OBS_bool, obs_region_is_include,
		     (const OBS_region *reg, const void *ptr))
{
	return reg->head &&
	    (const char *)reg->head <= (const char *)ptr &&
	    (const char *)ptr < (const char *)reg->head + reg->size;
}

/**
 * get pointer which is counted from lower side.
 *
 * @li obs_region_get_pointer_higher(reg, 0) is the same as reg->head.
 * @li obs_region_get_pointer_higher(reg, reg->size) points a byte just
 *     after the end of the region.
 *
 * @param reg		region.
 * @param pos		position from begin of the region, in a byte.
 */
OBS_DEF_INLINE_FUNC_(void *, obs_region_get_pointer_lower,
		     (const OBS_region *reg, OBS_size pos))
{ return (void *)((char *)reg->head + pos); }

/**
 * get pointer which is counted from higher side.
 *
 * @li obs_region_get_pointer_higher(reg, 0) points a byte just after
 *     the end of the region.
 * @li obs_region_get_pointer_higher(reg, reg->size) is the same as reg->head.
 *
 * @param reg		region.
 * @param pos		position from end of the region, in a byte.
 * @return		pointer which is specified by @a pos.
 */
OBS_DEF_INLINE_FUNC_(void *, obs_region_get_pointer_higher,
		     (const OBS_region *reg, OBS_size pos))
{ return obs_region_get_pointer_lower(reg, reg->size - pos); }

/**
 * get head of the region, same as reg->head.
 *
 * @param reg		region.
 * @return		pointer which points the head of @a reg.
 */
OBS_DEF_INLINE_FUNC_(void *, obs_region_begin, (const OBS_region *reg))
{ return reg->head; }

/**
 * get tail of the region.
 *
 * @param reg		region.
 * @return		pointer which points a byte just after the region.
 */
OBS_DEF_INLINE_FUNC_(void *, obs_region_end, (const OBS_region *reg))
{ return obs_region_get_pointer_higher(reg, 0); }

#endif
