/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file*/

#ifndef OBS_SLAB_H_
#define OBS_SLAB_H_

#include <obs/obs_pagealloc.h>

/**@name SlabFeatures
 * feature flags for slab allocator.
 * @{
 */

/** feature flag indicating to use large policy forcibly. */
#define OBS_SLAB_F_FORCE_LARGE_POLICY	0x0001
/**
 * feature flag indicating to prevent using hash table even when large policy.
 * In this case, each slab keeps only one object and the allocator may waste
 * memory.
 */
#define OBS_SLAB_F_FORCE_NO_HASH	0x0002
/**
 * Check whether a feature is applyed to a stack.
 *
 * @param slab		slab.
 * @param tag		feature name from which OBS_SLAB_F_ prefix is omitted.
 */
#define OBS_SLAB_IS_F(slab, tag)					      \
	(((slab)->features_ & OBS_SLAB_F_##tag) != 0)
/**
 * Check whether a feature is applyed to a mask.
 *
 * @param features	mask.
 * @param tag		feature name from which OBS_SLAB_F_ prefix is omitted.
 */
#define OBS_SLAB_FEATURES_IS_F(features, tag)				      \
	(((features) & OBS_SLAB_F_##tag) != 0)
/** @} */

/** slab allocator instance (opaque). */
typedef struct OBS_slab OBS_slab;
/** slab hooks (transparent). */
typedef struct OBS_slab_hooks OBS_slab_hooks;
/** @internal slab data. */
typedef struct OBS_slab_data_ OBS_slab_data_;
/** @internal slab operators. */
typedef struct OBS_slab_ops_ OBS_slab_ops_;

/**
 * hook function prototype.
 *
 * @param slab		slab instance.
 * @param hooks_priv	private data for hooks.
 * @param obj		current object.
 * @return		result status.
 */
typedef OBS_status (*OBS_slab_simple_hook_func)(
	OBS_slab *slab, void *hooks_priv, void *obj);

/**
 * slab hooks (transparent).
 *
 * Any hooks may be NULL.
 */
struct OBS_slab_hooks
{
	/** hook called to construct a new object. */
	OBS_slab_simple_hook_func construct;
	/** hook called to destruct an object. */
	OBS_slab_simple_hook_func destruct;
	/** hook called before the object is get. */
	OBS_slab_simple_hook_func get;
	/** hook called after the object is put back. */
	OBS_slab_simple_hook_func put;
};

/** @internal slab operator. */
struct OBS_slab_ops_
{
	/**
	 * uninit
	 * @param slab		slab.
	 */
	void (*uninit)(OBS_slab *slab);
	/**
	 * get an object.
	 * @param slab		slab.
	 * @param rptr		pointer to the pointer returned the object.
	 * @return		result status.
	 */
	OBS_status (*get)(OBS_slab *slab, void **rptr);
	/**
	 * put an object.
	 * @param slab		slab.
	 * @param obj		pointer to the object.
	 */
	void (*put)(OBS_slab *slab, void *obj);
	/**
	 * put all objects.
	 * @param slab		slab.
	 */
	void (*put_all)(OBS_slab *slab);
	/**
	 * run garbage collection.
	 * @param slab		slab.
	 * @param lowat		low water in number of objects, may be zero.
	 */
	void (*run_gc)(OBS_slab *slab, OBS_size lowat);
};

/** slab allocator instance (opaque). */
struct OBS_slab
{
	/**@internal page allocator instance. */
	OBS_pa pagealloc_;
	/**@internal object size. */
	OBS_size object_size_;
	/**@internal element size, which is object size + padding size. */
	OBS_size element_size_;
	/**@internal size of each slab. */
	OBS_size slab_size_;
	/**@internal number of slabs. */
	OBS_size num_slabs_;
	/**@internal wastage per a slab. */
	OBS_size wastage_per_slab_;
	/**@internal number of free objects. */
	OBS_size num_free_objects_;
	/**@internal number of objects per a slab. */
	OBS_size num_objects_per_slab_;
	/**@internal number of choises for top padding of a slab. */
	OBS_size num_ranks_of_top_;
	/**@internal high water. */
	OBS_size hiwat_;
	/**@internal low water. */
	OBS_size lowat_;
	/**@internal shift amount of alignment. */
	int alignment_;
	/**@internal pointer to hooks. */
	const OBS_slab_hooks *hooks_;
	/**@internal hooks' private data given by caller. */
	void *hooks_priv_;
	/**@internal slab ops. */
	const OBS_slab_ops_ *ops_;
	/**@internal private space for slab ops. */
	void *ops_priv_[8];
};

OBS_BEGIN_DECL_
/**
 * initialize slab with specifying page allocator class.
 *
 * @param slab		slab instance structure to be initialized.
 * @param objsize	objectsize.
 * @param features	feature mask.
 * @param align		shift amount of alignment.
 * @param hooks		hooks, may be NULL.
 * @param hooks_priv	private data for hooks.
 * @param paops		page allocator class.
 * @return		result status.
 */
OBS_status obs_slab_init_pa_ops(
	OBS_slab *slab, OBS_size objsize, OBS_flags features, int align,
	const OBS_slab_hooks *hooks, void *hooks_priv,
	const OBS_pa_ops *paops);
/**
 * uninitialize slab.
 *
 * @param slab		slab instance structure to be uninitialized.
 */
void obs_slab_uninit(OBS_slab *slab);
/**
 * @internal initialize small object slab.
 */
OBS_status obs_slab_small_init_(OBS_slab *, OBS_size, OBS_flags, int,
				const OBS_slab_hooks *, void *,
				const OBS_pa_ops *);
/**
 * @internal initialize large object slab.
 */
OBS_status obs_slab_large_init_(OBS_slab *, OBS_size, OBS_flags, int,
				const OBS_slab_hooks *, void *,
				const OBS_pa_ops *);
OBS_END_DECL_

/**
 * initialize slab with default page allocator class.
 *
 * @param slab		slab allocator instance structure to be initialized.
 * @param objsize	objectsize.
 * @param features	feature mask.
 * @param align		shift amount of alignment.
 * @param hooks		hooks, may be NULL.
 * @param hooks_priv	private data for hooks.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(
	OBS_status, obs_slab_init,
	(OBS_slab *slab, OBS_size objsize, OBS_flags features, int align,
	 const OBS_slab_hooks *hooks, void *hooks_priv))
{
	return obs_slab_init_pa_ops(slab, objsize, features, align,
				    hooks, hooks_priv,
				    obs_pa_get_current_ops());
}

/**
 * get an object which is initialized and unused.
 *
 * @param slab		slab allocator instance.
 * @param rptr		pointer to a pointer variable which is make pointing
 *			to the result object.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_slab_get, (OBS_slab *slab, void **rptr))
{
	return (*slab->ops_->get)(slab, rptr);
}

/**
 * put an object back to the allocator.
 *
 * @param slab		slab allocator instance.
 * @param ptr		pointer to the object to be put back.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(void, obs_slab_put, (OBS_slab *slab, void *ptr))
{
	(*slab->ops_->put)(slab, ptr);
}

/**
 * put all objects back which have been allocated until then.
 *
 * @param slab		slab allocator instance.
 */
OBS_DEF_INLINE_FUNC_(void, obs_slab_put_all, (OBS_slab *slab))
{
	(*slab->ops_->put_all)(slab);
}

/**
 * run garbage collection.
 *
 * free all unused slabs until the number of free objects falls below the
 * low water or there is no unused slab.
 *
 * @param slab		slab allocator instance.
 * @param lowat		low water.
 */
OBS_DEF_INLINE_FUNC_(void, obs_slab_run_gc, (OBS_slab *slab, OBS_size lowat))
{
	(*slab->ops_->run_gc)(slab, lowat);
}

/**
 * get per-slab wastage.
 *
 * @param slab		slab allocator instance to be examined.
 * @return		wastage per a slab, in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_slab_get_wastage_per_slab, (OBS_slab *slab))
{
	return slab->wastage_per_slab_;
}

/**
 * get slab size.
 *
 * @param slab		slab allocator instance to be examined.
 * @return		slab size in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_slab_get_slab_size, (OBS_slab *slab))
{
	return slab->slab_size_;
}

/**
 * get the current number of slabs which is keeped by the specified allocator.
 *
 * @param slab		slab allocator instance to be examined.
 * @return		number of slabs.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_slab_get_num_slabs, (OBS_slab *slab))
{
	return slab->num_slabs_;
}

/**
 * get the current low water mark.
 *
 * @param slab		slab allocator instance to be examined.
 * @return		number of objects as the low water mark.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_slab_get_low_water, (OBS_slab *slab))
{
	return slab->lowat_;
}

/**
 * get the current high water mark.
 *
 * @param slab		slab allocator instance to be examined.
 * @return		number of objects as the high water mark.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_slab_get_high_water, (OBS_slab *slab))
{
	return slab->hiwat_;
}

/**
 * set water marks.
 *
 * Whenever objects are put back to the allocator, the number of free objects
 * is checked.  If the number is more than the high water mark,
 * garbage collection is invoked.  Unused slabs are freed until the number of
 * free objects falls below the low water or there is no unused slab.
 *
 * @param slab		slab allocator instance to be examined.
 * @param lo		low water mark.
 * @param hi		high water mark.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_slab_set_water_mark,
		     (OBS_slab *slab, OBS_size lo, OBS_size hi))
{
	if (lo != 0 && lo >= hi)
		return OBS_S_INVAL;
	slab->lowat_ = lo;
	slab->hiwat_ = hi;
	return OBS_S_NOERR;
}

#endif
