/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs_stack.h>
#include <string.h>

OBS_DEF_INLINE_FUNC_(OBS_size, calc_blocks, (OBS_stack *st, OBS_size size))
{
	OBS_size blocksize = obs_stack_get_chunk_size(st), newsize;
	OBS_bool dbl = OBS_STACK_IS_F(st, EXPONENTIAL_EXPAND);

	if (dbl) {
		newsize = blocksize;
		while (newsize && newsize < size)
			newsize <<= 1;
		if (newsize == 0)
			return 0;
	} else {
		if (size == 0)
			newsize = blocksize;
		else {
			newsize = size + blocksize - 1;
			if (newsize < size)
				return 0;
		}
	}
	return newsize / blocksize;
}

static OBS_status
expand(OBS_stack *st, OBS_size desired_size, OBS_bool *isfrag)
{
	OBS_size blocks;
	OBS_region newreg, lastreg;
	OBS_status ret;
	OBS_stack_chunk_ *ch, *prev;

	if (isfrag)
		*isfrag = OBS_FALSE;

	blocks = calc_blocks(st, desired_size + sizeof (OBS_stack_chunk_));
	if (blocks == 0)
		return OBS_S_TOO_BIG;

	if (st->last_) {
		prev = st->last_->prev;
		obs_stack_chunk_to_block_(st->last_, &newreg);
		if (!OBS_STACK_IS_F(st, NO_CONCATENATE) &&
		    OBS_PA_IS_F(&st->pagealloc_, EXPAND_TAIL)) {
			OBS_size incr = blocks;
			if (st->top_.head)
				incr -=
				    ((OBS_uintptr)st->last_ -
				     (OBS_uintptr)st->top_.head) /
				    obs_stack_get_chunk_size(st);
			ret = obs_pa_expand_tail(
				&st->pagealloc_, &newreg, incr);
			if (ret == OBS_S_NOERR)
				goto expanded;
			else if (ret != OBS_S_NOTSUPP &&
				 ret != OBS_S_CANNOT_EXPAND)
				return ret;
		}
	}

	prev = st->last_;
	ret = obs_pa_alloc(&st->pagealloc_, &newreg, blocks);
	if (ret)
		return ret;
	if (st->last_ && !OBS_STACK_IS_F(st, NO_CONCATENATE)) {
		obs_stack_chunk_to_block_(st->last_, &lastreg);
		ret = obs_pa_concatenate(&st->pagealloc_, &lastreg, &newreg);
		if (ret == OBS_S_NOT_CONTIGUOUS ||
		    ret == OBS_S_NOTSUPP) {
			if (isfrag)
				*isfrag = OBS_TRUE;
		} else if (ret) {
			obs_pa_free(&st->pagealloc_, &newreg);
			return ret;
		} else { /* concatenation succeeded */
			newreg = lastreg;
			prev = prev->prev;
		}
	} else if (isfrag)
		*isfrag = OBS_TRUE;

expanded:
	ch = obs_stack_block_to_chunk_(&newreg);
	ch->size = obs_stack_block_to_chunk_size_(&newreg);
	ch->prev = prev;
	st->last_ = ch;

	return OBS_S_NOERR;
}

#define AWARE_PA_FEATURES (OBS_PA_F_EXPAND_TAIL |			      \
			   OBS_PA_F_SHRINK_HEAD |			      \
			   OBS_PA_F_SHRINK_TAIL |			      \
			   OBS_PA_F_ALL_FREE_ON_UNINIT |		      \
			   OBS_PA_F_CONCATENATE)

OBS_status
obs_stack_init_pa_ops(OBS_stack *st, OBS_size numunit, OBS_flags features,
		      int align, const OBS_pa_ops *paops)
{
	OBS_status ret;
	OBS_flags pafeat;
	int maxalign;

	pafeat = obs_pa_ops_get_available_features(paops);
	pafeat &= AWARE_PA_FEATURES;
	maxalign = obs_pa_ops_get_alignment(paops);
	if (align<0 || align > maxalign)
		return OBS_S_INVAL;

	ret = obs_pa_ops_init(paops, &st->pagealloc_, numunit, pafeat);
	if (ret)
		return ret;
	st->alignment_ = align;
	st->alignment_mask_ = (1 << align) - 1;
	st->last_ = NULL;
	st->features_ = features;
	obs_region_init(&st->top_);

	return OBS_S_NOERR;
}

void
obs_stack_uninit(OBS_stack *st)
{
	if (!OBS_PA_IS_F(&st->pagealloc_, ALL_FREE_ON_UNINIT))
		obs_stack_free_all(st);
	obs_pa_uninit(&st->pagealloc_);
	memset(st, 0, sizeof (OBS_stack));
}

OBS_status
obs_stack_truncate_after(OBS_stack *st, void *ptr)
{
	OBS_stack_chunk_ *curlast = st->last_;

	while (st->last_ != NULL &&
	       (ptr == NULL ||
		!obs_stack_is_pointer_in_chunk_(st->last_, ptr))) {
		OBS_stack_chunk_ *prev = st->last_->prev;
		OBS_region reg;
		obs_stack_chunk_to_block_(st->last_, &reg);
		obs_pa_free(&st->pagealloc_, &reg);
		st->last_ = prev;
	}
	if (st->last_ == NULL) {
		obs_region_init(&st->top_);
		if (ptr)
			return OBS_S_REVOKED;
		return OBS_S_NOERR;
	}

	if (st->last_ != curlast || (char *)st->top_.head > (char *)ptr) {
		ptr = (void *)obs_stack_round_up_(st, (OBS_uintptr)ptr);
		st->top_.head = ptr;
		st->top_.size = 0;
	} else if (obs_region_is_valid(&st->top_) &&
		   (char *)ptr - (char *)st->top_.head < st->top_.size)
		st->top_.size = (char *)ptr - (char *)st->top_.head;

	if (!OBS_STACK_IS_F(st, NO_SHRINK_AFTER) &&
	    OBS_PA_IS_F(&st->pagealloc_, SHRINK_TAIL)) {
		OBS_size blocksize = obs_stack_get_chunk_size(st);
		OBS_size desired_size =
		    (char *)ptr-(char *)obs_stack_get_chunk_begin_(st->last_) +
		    sizeof (OBS_stack_chunk_);
		OBS_size newsize = calc_blocks(st, desired_size) * blocksize;
		OBS_region reg;

		obs_stack_chunk_to_block_(st->last_, &reg);
		if (newsize < reg.size) {
			OBS_stack_chunk_ *prevch = st->last_->prev;
			OBS_status ret;

			ret = obs_pa_shrink_tail(&st->pagealloc_, &reg,
						 (reg.size-newsize)/blocksize);
			if (ret == OBS_S_NOERR) {
				OBS_stack_chunk_ *newch =
				    obs_stack_block_to_chunk_(&reg);
				newch->size =
				    obs_stack_block_to_chunk_size_(&reg);
				newch->prev = prevch;
				st->last_ = newch;
			}
		}
	}

	return OBS_S_NOERR;
}

OBS_status
obs_stack_truncate_before(OBS_stack *st, void *ptr)
{
	OBS_stack_chunk_ *ch, *prev, *curch;

	for (curch=st->last_;
	     curch && !obs_stack_is_pointer_in_chunk_(curch, ptr);
	     curch=curch->prev)
		;

	if (!curch)
		return ptr ? OBS_S_INVAL : OBS_S_NOERR;

	prev = curch->prev;
	curch->prev = NULL;
	while ((ch = prev) != NULL) {
		OBS_region reg;
		prev = ch->prev;
		obs_stack_chunk_to_block_(ch, &reg);
		obs_pa_free(&st->pagealloc_, &reg);
	}

	ptr = (void *)obs_stack_round_down_(st, (OBS_uintptr)ptr);
	if (st->last_ == curch) {
		if (!obs_region_is_valid(&st->top_) ||
		    (char *)st->top_.head + st->top_.size < (char *)ptr) {
			ptr = obs_stack_get_chunk_end_(curch);
			obs_region_init(&st->top_);
		} else if ((char *)st->top_.head < (char *)ptr) {
			st->top_.size -= (char *)ptr - (char *)st->top_.head;
			st->top_.head = ptr;
		}
	}

	if (!OBS_STACK_IS_F(st, NO_SHRINK_BEFORE) &&
	    OBS_PA_IS_F(&st->pagealloc_, SHRINK_HEAD)) {
		OBS_size blocksize = obs_stack_get_chunk_size(st);
		OBS_size desired_size =
		    (char *)obs_stack_get_chunk_end_(curch) - (char *)ptr +
		    sizeof (OBS_stack_chunk_);
		OBS_size newsize = calc_blocks(st, desired_size) * blocksize;
		OBS_region reg;

		obs_stack_chunk_to_block_(curch, &reg);
		if (newsize < reg.size) {
			OBS_status ret;

			ret = obs_pa_shrink_head(&st->pagealloc_, &reg,
						 (reg.size-newsize)/blocksize);
			if (ret == OBS_S_NOERR) {
				curch->size =
				    obs_stack_block_to_chunk_size_(&reg);
			}
		}
	}
	if (!obs_region_is_valid(&st->top_))
		st->top_.head = obs_stack_get_chunk_begin_(st->last_);

	return OBS_S_NOERR;
}

OBS_status
obs_stack_set_alignment(OBS_stack *st, int align)
{
	int maxalign;

	maxalign = obs_stack_get_max_alignment(st);
	if (align<1 || align > maxalign)
		return OBS_S_INVAL;

	if (st->top_.head)
		st->top_.head =
		    (void *)((char *)st->top_.head + st->top_.size);
	st->top_.size = 0;

	return OBS_S_NOERR;
}

static void
handle_orphaned_chunk(OBS_stack *st)
{
	OBS_status ret;
	OBS_region ophreg, curreg;
	OBS_stack_chunk_ *orphaned = st->last_->prev;
	OBS_stack_chunk_ *prev = orphaned->prev;

	/* ASSERT: st->last_->prev == orphaned */
	obs_stack_chunk_to_block_(st->last_, &curreg);
	obs_stack_chunk_to_block_(orphaned, &ophreg);

	ret = OBS_S_NOT_CONTIGUOUS;
	if (!OBS_STACK_IS_F(st, NO_CONCATENATE))
		ret = obs_pa_concatenate(&st->pagealloc_, &curreg, &ophreg);
	if (ret == OBS_S_NOERR) {
		/* merge */
		OBS_stack_chunk_ *ch;
		ch = obs_stack_block_to_chunk_(&curreg);
		ch->size = obs_stack_block_to_chunk_size_(&curreg);
		st->last_ = ch;
	} else /* free */
		obs_pa_free(&st->pagealloc_, &ophreg);
	st->last_->prev = prev;
}

static void
shrink_reloc(OBS_stack *st)
{
	OBS_stack_chunk_ *leaved;
	OBS_size desired_size, blocksize, newsize;
	OBS_region reg;

	if (st->top_.head == NULL ||
	    OBS_STACK_IS_F(st, NO_SHRINK_RELOC) ||
	    !OBS_PA_IS_F(&st->pagealloc_, SHRINK_TAIL))
		return;

	leaved = st->last_->prev;
	desired_size =
	    (char *)st->top_.head - (char *)obs_stack_get_chunk_begin_(leaved) +
	    sizeof (OBS_stack_chunk_);
	blocksize = obs_stack_get_chunk_size(st);
	newsize = desired_size + blocksize - 1;
	newsize -= newsize % blocksize;
	obs_stack_chunk_to_block_(leaved, &reg);
	if (newsize < reg.size) {
		OBS_stack_chunk_ *prevch = leaved->prev;
		OBS_status ret;

		ret = obs_pa_shrink_tail(&st->pagealloc_, &reg,
					 (reg.size-newsize)/blocksize);
		if (ret == OBS_S_NOERR) {
			leaved = obs_stack_block_to_chunk_(&reg);
			leaved->size = obs_stack_block_to_chunk_size_(&reg);
			leaved->prev = prevch;
			st->last_->prev = leaved;
		}
	}
}


OBS_status
obs_stack_reserve_(OBS_stack *st, OBS_size desired_size)
{
	OBS_bool isfrag;
	OBS_status ret;
	OBS_bool issole = obs_stack_is_sole_in_last_chunk_(st);

	ret = expand(st, desired_size, &isfrag);
	if (ret)
		return ret;
	if (isfrag) {
		void *newtop = obs_stack_get_chunk_begin_(st->last_);
		if (st->last_->prev) {
			if (st->top_.head)
				memcpy(newtop, st->top_.head, st->top_.size);
			if (issole)
				handle_orphaned_chunk(st);
			else
				shrink_reloc(st);
		}
		st->top_.head = newtop;
	}

	return OBS_S_NOERR;
}
