/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file */

#ifndef OBS_STACK_H_
#define OBS_STACK_H_

#include <obs/obs_pagealloc.h>
#include <string.h>

/**@name StackFeatures
 * feature flags for stack allocator.
 * @{
 */

/**
 * Feature flag indicating exponential expanse.
 * OBS tends to allocate memory blocks by power of 2
 * if this is specified.  This is advantageous in calculation order,
 * but make allocator tend to waste memory.
 * Otherwise, OBS expands memory blocks by linear.
 */
#define OBS_STACK_F_EXPONENTIAL_EXPAND	0x0001
/**
 * Feature flag indicating restrain concatenation of the blocks.
 * Memory blocks are left small pieces as if they can combine.
 * Normally this is unnecessary.
 */
#define OBS_STACK_F_NO_CONCATENATE	0x0002
/**
 * Feature flag indicating restrain shrinking memory blocks in
 * obs_stack_truncate_after().  Normally this is unnecessary.
 */
#define OBS_STACK_F_NO_SHRINK_AFTER	0x0004
/**
 * Feature flag indicating restrain shrinking memory blocks in
 * obs_stack_truncate_before().  Normally this is unnecessary.
 */
#define OBS_STACK_F_NO_SHRINK_BEFORE	0x0008
/**
 * Feature flag indicating restrain shrinking memory blocks at relocation.
 * Normally this is unnecessary.
 */
#define OBS_STACK_F_NO_SHRINK_RELOC	0x0010

/**
 * Check whether a feature is applyed to a stack.
 *
 * @param st		stack.
 * @param tag		feature name from which OBS_STACK_F_ prefix is omitted.
 */
#define OBS_STACK_IS_F(st, tag)						      \
	(((st)->features_ & OBS_STACK_F_##tag) != 0)
/**
 * Check whether a feature is applyed to a mask.
 *
 * @param features	mask.
 * @param tag		feature name from which OBS_STACK_F_ prefix is omitted.
 */
#define OBS_STACK_FEATURES_IS_F(features, tag)				      \
	(((features) & OBS_STACK_F_##tag) != 0)
/** @} */

/** stack instance. (opaque) */
typedef struct OBS_stack OBS_stack;
/** stack information. (transparent) */
typedef struct OBS_stack_info OBS_stack_info;
/** @internal stack chunk footer. */
typedef struct OBS_stack_chunk_ OBS_stack_chunk_;

/**
 * @internal stack chunk footer.
 *
 * Each stack is normally divided in multiple chunks.
 * This is placed in the highest position of each chunk.
 */
struct OBS_stack_chunk_
{
	/** previous chunk, further from stack top than this chunk. */
	OBS_stack_chunk_ *prev;
	/**
	 * data area size of this chunk.
	 * @a size + sizeof OBS_stack_chunk_ = block size
	 */
	OBS_size size;
};

/** stack allocator instance. (opaque) */
struct OBS_stack
{
	/**
	 * @internal page allocator instance which is used to allocate chunks.
	 */
	OBS_pa pagealloc_;
	/** @internal chunk including stack top. */
	OBS_stack_chunk_ *last_;
	/** @internal current feature mask. */
	OBS_flags features_;
	/** @internal region of the stack top object. */
	OBS_region top_;
	/** @internal shift amount of alignment.
	 * 0 means 1 byte alignment, 1 means 2 bytes, and so on.
	 */
	int alignment_;
	/** @internal alignment mask. */
	OBS_uintptr alignment_mask_;
};

/** stack information structure. (transparent) */
struct OBS_stack_info
{
	/** page unit size of the page allocator. */
	OBS_size page_unit_size;
	/** block size of each chunk. */
	OBS_size block_size;
	/**
	 * unavailable size of each chunk in a byte.
	 * typically same as the size of OBS_stack_chunk_.
	 */
	OBS_size unavailable_per_block;
	/** feature mask. */
	OBS_flags features;
	/** shift amount of alignment.
	 * 0 means 1 byte alignment, 1 means 2 bytes, and so on.
	 */
	int alignment;
};

OBS_BEGIN_DECL_
/**
 * initialize stack.
 *
 * @param stack		stack to be initialized.
 * @param num_pages	chunk size in a page.
 * @param features	feature mask.
 * @param align		shift amount of alignment value,
 *			restricted equal to or less than the value
 *                      obs_pa_get_alignment(@a paops) returns.
 * @param paops		page allocator class.
 * @return		result status.
 */
OBS_status obs_stack_init_pa_ops(
	OBS_stack *stack, OBS_size num_pages, OBS_flags features,
	int align, const OBS_pa_ops *paops);

/**
 * uninitialize stack
 *
 * @param stack		stack to be uninitialized.
 */
void obs_stack_uninit(OBS_stack *stack);

/**
 * truncate stack from specified object to the last.
 *
 * @param stack		stack.
 * @param obj		object.
 * @return		result status.
 */
OBS_status obs_stack_truncate_after(OBS_stack *stack, void *obj);

/**
 * truncate stack from the first to object just before specified object.
 *
 * @param stack		stack.
 * @param obj		object.
 * @return		result status.
 */
OBS_status obs_stack_truncate_before(OBS_stack *stack, void *obj);

/**
 * change alignment of stack.
 *
 * @param stack		stack.
 * @param align		shift amount of alignment value.
 * @return		result status.
 * @note		this will revoke non-fixated stack top object.
 */
OBS_status obs_stack_set_alignment(OBS_stack *stack, int align);

/**@internal
 * reserve additional bytes for the current non-fixated stack top.
 *
 * @param stack		stack.
 * @param size		size to be reserved in a byte.
 * @return		result status.
 */
OBS_status obs_stack_reserve_(OBS_stack *stack, OBS_size size);
OBS_END_DECL_

/**
 * get page unit size of specified page allocator class.
 *
 * @param ops		page allocator class object.
 * @return		page size.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_get_page_unit_size_pa_ops,
		     (const OBS_pa_ops *ops))
{ return obs_pa_ops_get_page_unit(ops); }

/**
 * get page unit size fo default page allocator class.
 *
 * @return		page size.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_get_page_unit_size, (void))
{ return obs_stack_get_page_unit_size_pa_ops(obs_pa_get_current_ops()); }

/**
 * unavailable size of each chunk for user data.
 *
 * unavailable portion is reserved for the management area of the allocator.
 * this value may be used to determine the best chunk size.
 *
 * @param features	feature mask.
 * @return		unavailable size in a byte.
 */
OBS_DEF_INLINE_FUNC_AU_(OBS_size, obs_stack_get_unavailable_per_block,
			  (OBS_flags features))
{ return sizeof (OBS_stack_chunk_); }

/**
 * get information of the stack.
 *
 * @param st		stack.
 * @param info		pointer to the OBS_stack_info structure to which
 *                      the information is stored.
 */
OBS_DEF_INLINE_FUNC_(void, obs_stack_get_info,
		     (const OBS_stack *st, OBS_stack_info *info))
{
	info->page_unit_size =
	    obs_pa_ops_get_page_unit(obs_pa_get_ops(&st->pagealloc_));
	info->block_size = obs_pa_get_block_size(&st->pagealloc_);
	info->unavailable_per_block = sizeof (OBS_stack_chunk_);
	info->features = st->features_;
	info->alignment = st->alignment_;
}

/**
 * free all objects in the stack.
 *
 * @param st		stack.
 */
OBS_DEF_INLINE_FUNC_(void, obs_stack_free_all, (OBS_stack *st))
{ obs_stack_truncate_after(st, NULL); }

/**
 * revoke non-fixated stack top object.
 *
 * @param st		stack.
 */
OBS_DEF_INLINE_FUNC_(void, obs_stack_revoke_top, (OBS_stack *st))
{ st->top_.size = 0; }

/**
 * initialize stack with using the default page allocator.
 * @param stack		stack to be initialized.
 * @param num_pages	chunk size in a page.
 * @param features	feature mask.
 * @param alignment	shift amount of alignment value,
 *			restricted equal to or less than the value
 *                      obs_pa_get_alignment(@a paops) returns.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_init,
		     (OBS_stack *stack, OBS_size num_pages, OBS_flags features,
		      int alignment))
{
	return obs_stack_init_pa_ops(
		stack, num_pages, features, alignment,
		obs_pa_get_current_ops());
}

/**
 * get current feature mask of the stack.
 *
 * @param st		stack.
 * @return		feature mask.
 */
OBS_DEF_INLINE_FUNC_(OBS_flags, obs_stack_get_features,
		     (const OBS_stack *st))
{ return st->features_; }

/**
 * get current non-fixated stack top object.
 *
 * @param st		stack.
 * @return		pointer to the region structure indicating
 *			the stack top object.
 */
OBS_DEF_INLINE_FUNC_(const OBS_region *, obs_stack_get_top,
		     (const OBS_stack *st))
{ return &st->top_; }

/**
 * check whether the current stack top object is valid.
 *
 * @param st		stack.
 * @return		true if the current stack top is valid,
 *			otherwise false.
 */
OBS_DEF_INLINE_FUNC_(OBS_bool, obs_stack_is_top_valid,
		     (const OBS_stack *st))
{ return st->top_.head != NULL && st->top_.size != 0; }

/**
 * get chunk size of the stack.
 *
 * @param st		stack.
 * @return		chunk size in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_get_chunk_size,
		     (const OBS_stack *st))
{ return obs_pa_get_block_size(&st->pagealloc_); }

/**
 * get shift amount of max alignment size of the stack.
 *
 * @param st		stack.
 * @return		shift amount of the alignment value.
 */
OBS_DEF_INLINE_FUNC_(int, obs_stack_get_max_alignment, (const OBS_stack *st))
{ return obs_pa_ops_get_alignment(obs_pa_get_ops(&st->pagealloc_)); }

/**
 * get shift amount of the current alignment size of the stack.
 *
 * @param st		stack.
 * @return		shift amount of the alignment value.
 */
OBS_DEF_INLINE_FUNC_(int, obs_stack_get_alignment, (const OBS_stack *st))
{ return st->alignment_; }

/**@internal
 * get the pointer to the chunk head.
 *
 * @param ch		chunk footer.
 * @return		chunk head.
 */
OBS_DEF_INLINE_FUNC_(void *, obs_stack_get_chunk_begin_,
		     (OBS_stack_chunk_ *ch))
{
	return (void *)((char *)(void *)ch - ch->size);
}

/**@internal
 * get the pointer to the byte just after the user area of the chunk.
 *
 * @param ch		chunk footer.
 * @return		chunk end.
 */
OBS_DEF_INLINE_FUNC_(void *, obs_stack_get_chunk_end_, (OBS_stack_chunk_ *ch))
{
	return (void *)ch;
}

/**@internal
 * convert chunk footer to the corresponding block.
 *
 * @param ch		chunk footer.
 * @param reg		region structure to which the block is stored.
 */
OBS_DEF_INLINE_FUNC_(void, obs_stack_chunk_to_block_,
		     (OBS_stack_chunk_ *ch, OBS_region *reg))
{
	reg->head = obs_stack_get_chunk_begin_(ch);
	reg->size = ch->size + sizeof (OBS_stack_chunk_);
}

/**@internal
 * convert raw block to the chunk footer.
 *
 * @param reg		raw block.
 * @return		chunk footer.
 */
OBS_DEF_INLINE_FUNC_(OBS_stack_chunk_ *, obs_stack_block_to_chunk_,
		     (const OBS_region *reg))
{
	return obs_region_get_pointer_higher(reg, sizeof (OBS_stack_chunk_));
}

/**@internal
 * get user data area corresponding to the specified chunk footer.
 *
 * @param ch		chunk footer.
 * @param reg		region structure to which the area is stored.
 */
OBS_DEF_INLINE_FUNC_(void, obs_stack_chunk_to_data_,
		     (OBS_stack_chunk_ *ch, OBS_region *reg))
{
	reg->head = obs_stack_get_chunk_begin_(ch);
	reg->size = ch->size;
}

/**@internal
 * get size of user data area corresponding to the specified block.
 *
 * @param reg		raw block.
 * @return		size of user data area in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_block_to_chunk_size_,
		     (const OBS_region *reg))
{
	return reg->size - sizeof (OBS_stack_chunk_);
}

/**@internal
 * check whether the pointer is in the user data area corresponding to
 * the specified chunk.
 *
 * @param ch		chunk footer.
 * @param ptr		pointer to be examined.
 * @return		true if the pointer in the chunk, otherwise false.
 */
OBS_DEF_INLINE_FUNC_(OBS_bool, obs_stack_is_pointer_in_chunk_,
		     (OBS_stack_chunk_ *ch, void *ptr))
{
	return
	    (char *)obs_stack_get_chunk_begin_(ch) <= (char *)ptr &&
	    (char *)ptr < (char *)obs_stack_get_chunk_end_(ch);
}

/**@internal
 * round down the pointer by the alignment of the stack.
 *
 * @param st		stack.
 * @param ptr		pointer to be rounded down.
 * @return		rounded pointer.
 */
OBS_DEF_INLINE_FUNC_(OBS_uintptr, obs_stack_round_down_,
		     (const OBS_stack *st, OBS_uintptr ptr))
{
	return ptr & ~st->alignment_mask_;
}

/**@internal
 * round up the pointer by the alignment of the stack.
 *
 * @param st		stack.
 * @param ptr		pointer to be rounded up.
 * @return		rounded pointer.
 */
OBS_DEF_INLINE_FUNC_(OBS_uintptr, obs_stack_round_up_,
		     (const OBS_stack *st, OBS_uintptr ptr))
{
	return (ptr + st->alignment_mask_) & ~st->alignment_mask_;
}

/**@internal
 * get the aligned pointer just after the current non-fixated stack top.
 *
 * @param st		stack.
 * @return		result pointer.
 */
OBS_DEF_INLINE_FUNC_(OBS_uintptr, obs_stack_get_aligned_next_,
		     (OBS_stack *st))
{
	return obs_stack_round_up_(
		st, (OBS_uintptr)st->top_.head + st->top_.size);
}

/**@internal
 * check whether the current non-fixated stack top is sole in the last chunk.
 *
 * @param st		stack.
 * @return		true if the stack top is sole, otherwise false.
 */
OBS_DEF_INLINE_FUNC_(OBS_bool, obs_stack_is_sole_in_last_chunk_,
		     (OBS_stack *st))
{
	return
	    st->last_ &&
	    st->top_.head == obs_stack_get_chunk_begin_(st->last_);
}

/**
 * get free size of the current last chunk.
 *
 * @param st		stack.
 * @return		free size of the last chunk in a byte.
 */
OBS_DEF_INLINE_FUNC_(OBS_size, obs_stack_get_free_size, (OBS_stack *st))
{
	if (!st->last_)
		return 0;
	else if (!obs_region_is_valid(&st->top_))
		return st->last_->size;

	return
	    (char *)obs_stack_get_chunk_end_(st->last_) -
	    (char *)obs_region_end(&st->top_);
}

/**
 * reserve free area for the current non-fixated stack top object.
 * this may relocate the stack top, but do not change its size.
 * this is useful when the final object size is roughly known.
 *
 * @param st		stack.
 * @param size		reserved size in a byte, including already allocated
 *			amount for the stack top object (not increment size).
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_reserve,
		     (OBS_stack *st, OBS_size size))
{
	OBS_uintptr p;

	if (!st->last_)
		return obs_stack_reserve_(st, size);

	if (!obs_region_is_valid(&st->top_))
		st->top_.head = obs_stack_get_chunk_begin_(st->last_);
	else if (st->top_.size == 0)
		st->top_.head = (void *)obs_stack_get_aligned_next_(st);

	if (st->top_.size >= size)
		return OBS_S_NOERR;

	p = (OBS_uintptr)st->top_.head;
	if (p + size < p || /* overflow */
	    p + size > (OBS_uintptr)obs_stack_get_chunk_end_(st->last_))
		return obs_stack_reserve_(st, size);

	return OBS_S_NOERR;
}

/**
 * allocate a new non-fixated object.
 * this fixates the last non-fixated object.
 * because the area information of the fixated object,
 * the caller will need to save it before call this function.
 *
 * @param st		stack.
 * @param size		size of a new object in a byte.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_alloc,
		     (OBS_stack *st, OBS_size size))
{
	OBS_status ret;

	st->top_.head = (void *)obs_stack_get_aligned_next_(st);
	st->top_.size = 0;
	ret = obs_stack_reserve(st, size);
	if (ret)
		return ret;
	st->top_.size = size;

	return OBS_S_NOERR;
}

/**
 * fixate the current stack top object.
 *
 * @param st		stack.
 * @param reg		result memory region of the fixated stack top object.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_fixate_top,
		     (OBS_stack *st, OBS_region *reg))
{
	if (reg)
		*reg = *obs_stack_get_top(st);
	return obs_stack_alloc(st, 0);
}

/**
 * resize the current non-fixated stack top object.
 * this may relocate the stack top object.
 *
 * @param st		stack.
 * @param size		size after resized in a byte.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_resize,
		     (OBS_stack *st, OBS_size size))
{
	OBS_status ret;

	if (st->top_.size == 0)
		return obs_stack_alloc(st, size);

	ret = obs_stack_reserve(st, size);
	if (ret)
		return ret;

	st->top_.size = size;

	return OBS_S_NOERR;
}

/**
 * expand the current non-fixated stack top object.
 * this may relocate the stack top object.
 *
 * @param st		stack.
 * @param incr		increnemt in a byte.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_expand,
		     (OBS_stack *st, OBS_size incr))
{
	if (st->top_.size == 0)
		return obs_stack_alloc(st, incr);
	if (st->top_.size + incr < st->top_.size)
		return OBS_S_INVAL;
	return obs_stack_resize(st, st->top_.size + incr);
}

/**
 * expand the current non-fixed stack top object and append the data.
 * this may relocate the stack top object.
 *
 * @param st		stack.
 * @param data		pointer to the data to be copied.
 * @param size		data size.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append,
		     (OBS_stack *st, const void *data, OBS_size size))
{
	OBS_status ret = obs_stack_expand(st, size);
	if (ret)
		return ret;
	memcpy(obs_region_get_pointer_higher(&st->top_, size), data, size);
	return OBS_S_NOERR;
}

/**
 * expand the current non-fixed stack top object and append the data.
 * this may relocate the stack top object.
 *
 * @param st		stack.
 * @param reg		region to be copied.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_region,
		     (OBS_stack *st, const OBS_region *reg))
{ return obs_stack_append(st, reg->head, reg->size); }

/**
 * expand the current non-fixed stack top object and append a byte.
 * this may relocate the stack top object.
 *
 * @param st		stack.
 * @param ch		byte to be copied.
 * @return		result status.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_1,
		     (OBS_stack *st, char ch))
{
	OBS_status ret = obs_stack_expand(st, 1);
	if (ret)
		return ret;
	*(char *)obs_region_get_pointer_higher(&st->top_, 1) = ch;
	return OBS_S_NOERR;
}

/**
 * same as obs_stack_resize, but this assumes the free size of chunk is
 * sufficient thus omits checking the size.
 *
 * @param st		stack.
 * @param size		size after resized in a byte.
 * @return		result status, always OBS_S_NOERR.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_resize_fast,
		     (OBS_stack *st, OBS_size size))
{ st->top_.size = size; return OBS_S_NOERR; }

/**
 * same as obs_stack_expand, but this assumes the free size of chunk is
 * sufficient thus omits checking the size.
 *
 * @param st		stack.
 * @param incr		increnemt in a byte.
 * @return		result status, always OBS_S_NOERR.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_expand_fast,
		     (OBS_stack *st, OBS_size incr))
{ st->top_.size += incr; return OBS_S_NOERR; }

/**
 * same as obs_stack_append, but this assumes the free size of chunk is
 * sufficient thus omits checking the size.
 *
 * @param st		stack.
 * @param data		pointer to the data to be copied.
 * @param size		data size.
 * @return		result status, always OBS_S_NOERR.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_fast,
		     (OBS_stack *st, const void *data, OBS_size size))
{
	memcpy(obs_region_get_pointer_lower(
		       &st->top_, st->top_.size), data, size);
	st->top_.size += size;
	return OBS_S_NOERR;
}

/**
 * same as obs_stack_append_region, but this assumes the free size
 * of chunk is sufficient thus omits checking the size.
 *
 * @param st		stack.
 * @param reg		region to be copied.
 * @return		result status, always OBS_S_NOERR.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_region_fast,
		     (OBS_stack *st, const OBS_region *reg))
{ return obs_stack_append_fast(st, reg->head, reg->size); }

/**
 * same as obs_stack_append_1, but this assumes the free size
 * of chunk is sufficient thus omits checking the size.
 *
 * @param st		stack.
 * @param ch		byte to be copied.
 * @return		result status, always OBS_S_NOERR.
 */
OBS_DEF_INLINE_FUNC_(OBS_status, obs_stack_append_1_fast,
		     (OBS_stack *st, char ch))
{
	*(char *)obs_region_get_pointer_lower(&st->top_, st->top_.size) = ch;
	st->top_.size++;
	return OBS_S_NOERR;
}

#endif
