#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <obs/obs.h>

void
error_stdc(const char *msg, int err)
{
	fprintf(stderr, "%s: %s\n", msg, strerror(err));
	exit(1);
}

void
error_obs(const char *msg, OBS_status err)
{
	fprintf(stderr, "%s: %s\n", msg, obs_strstatus(err));
	exit(1);
}

OBS_bool isz = OBS_FALSE;
OBS_bool ise = OBS_FALSE;

void
test_malloc(size_t size, int count)
{
	int i;
	void *ptr;
	for (i=0; i<count; i++) {
		ptr = malloc(size);
		if (ptr == NULL)
			error_stdc("malloc", errno);
		if (isz)
			memset(ptr, 0, size);
	}
}

void
test_realloc(size_t size, int count)
{
	int i;
	void *ptr = NULL;
	size_t total = 0;
	for (i=0; i<count; i++) {
		if (total + size < total)
			error_stdc("overflow", EINVAL);
		total += size;
		ptr = realloc(ptr, total);
		if (ptr == NULL)
			error_stdc("realloc", errno);
	}
	if (isz)
		memset(ptr, 0, total);
	printf("realloc done : %p, %u\n", ptr, (unsigned)total);
}

void
test_obs_alloc(size_t size, int count)
{
	int i;
	OBS_status ret;
	OBS_stack stack;
	OBS_flags features = 0;

	if (ise)
		features |= OBS_STACK_F_EXPONENTIAL_EXPAND;

	ret = obs_stack_init(&stack, 1, features, OBS_ALIGN_1);
	if (ret)
		error_obs("obs_stack_init", ret);
	for (i=0; i<count; i++) {
		ret = obs_stack_alloc(&stack, size);
		if (ret)
			error_obs("obs_stack_alloc", ret);
		if (isz)
			memset(obs_stack_get_top(&stack)->head, 0,
			       obs_stack_get_top(&stack)->size);
	}
	obs_stack_uninit(&stack);
}

void
test_obs_resize(size_t size, int count)
{
	int i;
	OBS_status ret;
	OBS_stack stack;
	OBS_flags features = 0;

	if (ise)
		features |= OBS_STACK_F_EXPONENTIAL_EXPAND;

	ret = obs_stack_init(&stack, 1, features, OBS_ALIGN_1);
	if (ret)
		error_obs("obs_stack_init", ret);
	ret = obs_stack_alloc(&stack, 0);
	if (ret)
		error_obs("obs_stack_alloc", ret);
	for (i=0; i<count; i++) {
		ret = obs_stack_expand(&stack, size);
		if (ret)
			error_obs("obs_stack_expand", ret);
	}
	if (isz)
		memset(obs_stack_get_top(&stack)->head, 0,
		       obs_stack_get_top(&stack)->size);
	printf("expand done : %p, %u\n",
	       obs_stack_get_top(&stack)->head,
	       (unsigned)obs_stack_get_top(&stack)->size);
	obs_stack_uninit(&stack);
}
void
usage(void)
{
	fprintf(stderr,
		"usage: test_compare {-m|-o} [-r] [-z] <unit> <count>\n");
	exit(1);
}

int
main(int argc, char **argv)
{
	size_t size;
	int count;
	OBS_bool ism = OBS_FALSE, iso = OBS_FALSE, isr = OBS_FALSE;

	argv++; argc--;

	while (argv[0] && argv[0][0] == '-') {
		switch (argv[0][1]) {
		case 'm':
			ism = OBS_TRUE;
			break;
		case 'o':
			iso = OBS_TRUE;
			break;
		case 'r':
			isr = OBS_TRUE;
			break;
		case 'e':
			ise = OBS_TRUE;
			break;
		case 'z':
			isz = OBS_TRUE;
			break;
		default:
			usage();
		}
		argc--; argv++;
	}

	if (argc != 2)
		usage();

	size = atoi(argv[0]);
	count = atoi(argv[1]);

	if ((ism && iso) || (!ism && !iso) || size <= 0 || count <= 0)
		usage();

	if (ism) {
		if (ise)
			fprintf(stderr, "warning: -e has no effect.\n");
		if (isr)
			test_realloc(size, count);
		else
			test_malloc(size, count);
	} else {
		if (isr)
			test_obs_resize(size, count);
		else
			test_obs_alloc(size, count);
	}

	return 0;
}
